#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements interfaces with gds files.
"""


# TODO: Implement import functions (similar to load in python-gdsii)

import logging
import pprint

import numpy as np

import laygo2.object


#export functions
def export_from_laygo(db, filename, cellname=None, scale = 1e-9, layermapfile="default.layermap",
                      physical_unit=1e-9, logical_unit=0.001, pin_label_height=0.1,
                      pin_annotate_layer=['text', 'drawing'], text_height=0.1,
                      abstract_instances = False, abstract_instances_layer = ['prBoundary', 'drawing']):
    """
    Export specified laygo2 object(s) to a GDS file

    Parameters
    ----------
    db : laygo.object.database.Library
        a library object that designs to be exported.
    filename : str
        the name of the output file.
    cellname : list or str or None.
        the name of cells to be exported. If None, all cells in the libname are exported.
    scale : float
        the scaling factor that converts integer coordinates to actual ones (mostly 1e-6, to convert 1 to 1um).
    layermapfile : str
        the name of layermap file.
    physical_unit : float, optional
        GDS physical unit.
    logical_unit : float, optional
        GDS logical unit.
    pin_label_height : float, optional
        the height of pin label.
    pin_annotate_layer : [str, str], optional
        pin annotate layer name (used when pinname is different from netname).
    text_height : float, optional
        the height of text
    """
    raise Exception("GDS support is currently disabled. Will be reinvented by using gds-tk.")


def export(db, filename, cellname=None, scale = 1e-9, layermapfile="default.layermap",
           physical_unit=1e-9, logical_unit=0.001, pin_label_height=0.1,
           pin_annotate_layer=['text', 'drawing'], text_height=0.1,
           abstract_instances=False, abstract_instances_layer=['prBoundary', 'drawing']):
    """See laygo2.interface.gds.export_from_laygo for details."""
    export_from_laygo(db, filename, cellname, scale, layermapfile,
                      physical_unit, logical_unit, pin_label_height,
                      pin_annotate_layer, text_height,
                      abstract_instances, abstract_instances_layer)





import pytest
import numpy as np
import pprint
import laygo2

import test_tech as tech


def bb(a: int, b: int, c: int, d: int):
    return (np.asarray([[a, b], [c, d]]))

def test_grid_OneDimGrid():
    """
    DUT     : laygo2.object.grid.OneDimGrid
    Comment : test conversion operator
    Case    :  <, >, <=, >=
    """

    g1_o  = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 100], elements=[0, 10, 20, 40, 50])
    g1_no = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 200], elements=[10, 20, 25] )
    g1_e0 = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 180], elements=[0] )
    print("Testing 1d grid")

    test_set_index=[
        [ g1_o[:9],  np.array([0, 10, 20, 40, 50, 100, 110, 120, 140])],
        [ g1_o[1:] ,  np.array([10, 20, 40, 50]) ],
        [ g1_o[-6:10], np.array([-150, -100, -90, -80, -60, -50, 0, 10, 20, 40, 50, 100, 110, 120, 140, 150])],
        [ g1_o[ [3, 5, 7]], np.array([40, 100, 120]) ],

        [g1_no[:10], np.array([10, 20, 25, 210, 220, 225, 410, 420, 425,610])],
        [g1_no[1:], np.array([20, 25])],
        [g1_no[-6:10], np.array([
            -390, -380,-375,-190, -180, -175,
            10, 20, 25, 210, 220, 225, 410, 420, 425, 610
        ])],
        [g1_no[[3,5,7]], np.array([ 210,225, 420 ])],

        [g1_e0[:10], np.array([0, 180, 360, 540, 720, 900, 1080, 1260, 1440,1620])],
        [g1_e0[1:], np.array([180])],
        [g1_e0[-6:10], np.array([-1080, -900 ,-720 ,-540, -360,-180,0 , 180, 360, 540, 720, 900, 1080, 1260, 1440, 1620])],
        [g1_e0[ [3,5,7]], np.array([ 540, 900, 1260 ])],

    ]

    test_set_op_o=[
        # matched
        [g1_o == [-100, -80, -60, 0, 10, 40, 100, 110, 140], np.array([-5,-3,-2,0,1,3,5,6,8])],
        [g1_o < -60, -3], [g1_o < 0, -1], [g1_o < 120, 6],
        [g1_o < -100, -6], [g1_o < 100, 4],


        [g1_o <= -60, -2], [g1_o <= 0, 0], [g1_o <= 120, 7],
        [g1_o <= -100, -5], [g1_o <= 100, 5],

        [g1_o > -60, -1], [g1_o > 0, 1], [g1_o > 120, 8],
        [g1_o > -100, -4], [g1_o > 100, 6],

        [g1_o >= -60, -2], [g1_o >= 0, 0], [g1_o >= 120, 7],
        [g1_o >= -100,-5 ], [g1_o >= 100, 5],

        # between
        [g1_o == [-95, -75, 5, 15, 35, 105], np.array([None,None,None,None,None,None])],
        [g1_o < -95, -5], [g1_o < 45, 3], [g1_o < 105, 5],
        [g1_o <= -95, -5], [g1_o <= 45, 3], [g1_o <= 105, 5],
        [g1_o > -95, -4], [g1_o > 45, 4], [g1_o > 105, 6],
        [g1_o >= -95, -4], [g1_o >= 45, 4], [g1_o >= 105, 6],

        # upper out
        [g1_o == [-40, 60], np.array([None, None])],
        [g1_o < -40, -1], [g1_o < 60, 4],
        [g1_o <= -40, -1], [g1_o <= 60, 4],
        [g1_o > -40, 0], [g1_o > 60, 5],
        [g1_o >= -40, 0], [g1_o >= 60, 5],
        ]
    test_set_op_no = [
        # matched
        [g1_no == [-190, -180, -175, 10, 20, 25, 210, 220, 225], np.array([-3,-2,-1,0,1,2,3,4,5])],
        [g1_no < -190, -4], [g1_no < -175, -2], [g1_no < 10, -1], [g1_no < 25, 1],
        [g1_no < 210, 2], [g1_no < 225, 4],

        [g1_no <= -190, -3], [g1_no <= -175, -1], [g1_no <= 10, 0], [g1_no <= 25, 2],
        [g1_no <= 210, 3], [g1_no <= 225, 5],

        [g1_no > -190, -2], [g1_no > -175, 0], [g1_no > 10, 1], [g1_no > 25, 3],
        [g1_no > 210, 4], [g1_no > 225, 6],

        [g1_no >= -190, -3], [g1_no >= -175, -1], [g1_no >= 10, 0], [g1_no >= 25, 2],
        [g1_no >= 210, 3], [g1_no >= 225, 5],

        # between
        [g1_no == [-195, -185, -170, 15, 26, 30, 215], np.array([ None, None, None, None, None, None, None ])],
        [g1_no < -185, -3], [g1_no < 15, 0], [g1_no < 215, 3],
        [g1_no <= -185, -3], [g1_no <= 15, 0], [g1_no <= 215, 3],
        [g1_no > -185, -2], [g1_no > 15, 1], [g1_no > 215, 4],
        [g1_no >= -185, -2], [g1_no >= 15, 1], [g1_no >= 215, 4],

        # upper out
        [g1_no == [ -40,190], np.array([None, None])],
        [g1_no < -40, -1], [g1_no < 190, 2],
        [g1_no <= -40, -1], [g1_no <= 190, 2],
        [g1_no > -40, 0], [g1_no > 190, 3],
        [g1_no >= -40, 0], [g1_no >= 190, 3],

        # lower out
        [g1_no == [5, 205], np.array([None, None])],
        [g1_no < 5, -1], [g1_no < 205, 2],
        [g1_no <= 5, -1], [g1_no <= 205, 2],
        [g1_no > 5, 0], [g1_no > 205, 3],
        [g1_no >= 5, 0], [g1_no >= 205, 3],
    ]
    test_set_op_e0=[
        # matched
        [g1_e0 == [-180,0,180], np.array([ -1,0,1])],
        [g1_e0 <  -180,  -2], [g1_e0 <  0, -1], [g1_e0 < 180, 0],
        [g1_e0 <= -180, -1], [g1_e0 <= 0, 0],   [g1_e0 <= 180, 1],

        [g1_e0 >  -180,   0], [g1_e0 >  0, 1], [g1_e0 > 180, 2],
        [g1_e0 >= -180, -1], [g1_e0 >= 0, 0],   [g1_e0 >= 180, 1],

        [g1_e0 == [-170, 10, 170,190], np.array([None,None,None,None])],
        [g1_e0 < -170, -1], [g1_e0 < -10, -1], [g1_e0 < 170, 0], [g1_e0 < 190, 1],
        [g1_e0 <= -170, -1],[g1_e0 <= -10, -1], [g1_e0 <= 170, 0], [g1_e0 <= 190, 1],
        [g1_e0 > -170, 0],  [g1_e0 > -10, 0], [g1_e0 > 170, 1],[g1_e0 > 190, 2],
        [g1_e0 >= -170, 0], [g1_e0 >= -10, 0], [g1_e0 >= 170, 1],[g1_e0 >= 190, 2],

    ]


    i=0
    for q, a in test_set_index:
        i = i+1
        assert  np.all(q == a) == True , str("index ") + str(q) + str("  ") +  str(a) + str("  ") + str(i)
    i = 0
    for q, a in test_set_op_o:
        i = i + 1
        assert np.all(q == a) == True, str("ori ") + str(q) + str("  ") + str(a) + str("  ") + str(i)
    i = 0
    for q, a in test_set_op_no:
        i = i + 1
        assert np.all(q == a) == True, str("no ") + str(q) + str("  ") + str(a) + str("  ") + str(i)
    i = 0
    for q, a in test_set_op_e0:
        i = i + 1
        assert np.all(q == a) == True, str("e0 ") + str(q) + str("  ") + str(a) + str("  ") + str(i)

def test_grid_OneDimGrid_compare():
    """
    DUT     : laygo2.object.grid.OneDimGrid
    Comment : compare old conversion operator function and new
    Case    :  <, >, <=, >=
    """
    # compare older function vs new
    def _lt_1d(other, elements, width, shape):
        if isinstance(other, (int, np.integer)):
            quo = None
            mod = None
            for i, e in np.ndenumerate(elements):
                _quo = int(round(np.floor((other - e) / width)))
                if not (_quo * width + e == other):
                    if quo is None:
                        quo = _quo
                        mod = i[0]
                    elif _quo >= quo:  # update quo
                        quo = _quo
                        mod = i[0]
            return quo * shape + mod
        else:
            pass

    def _le_1d(other, elements, width, shape):
        if isinstance(other, (int, np.integer)):
            quo = None
            mod = None
            for i, e in np.ndenumerate(elements):
                _quo = int(round(np.floor((other - e) / width)))
                if quo is None:
                    quo = _quo
                    mod = i[0]
                elif _quo >= quo:  # update quo
                    quo = _quo
                    mod = i[0]
            return quo * shape + mod
        else:
            pass

    def _gt_1d(other, elements, width, shape):
        if isinstance(other, (int, np.integer)):
            quo = None
            mod = None
            for i, e in np.ndenumerate(elements):
                _quo = int(round(np.ceil((other - e) / width)))
                if not (_quo * width + e == other):
                    if quo is None:
                        quo = _quo
                        mod = i[0]
                    elif _quo < quo:  # update quo
                        quo = _quo
                        mod = i[0]
            return quo * shape + mod
        else:
            pass

    def _ge_1d(other, elements, width, shape):
        if isinstance(other, (int, np.integer)):
            quo = None
            mod = None
            for i, e in np.ndenumerate(elements):
                _quo = int(round(np.ceil((other - e) / width)))
                if quo is None:
                    quo = _quo
                    mod = i[0]
                elif _quo < quo:  # update quo
                    quo = _quo
                    mod = i[0]
            return quo * shape + mod
        else:
            pass

    test_set = [[[0, 35, 85, 130], 180, 4],
                [[0, 35, 85, 130], 140, 4],
                [[10, 35, 85, 130], 140, 4],
                [[0, 36, 86, 132], 140, 4],
                ]


    test_data = range(-1000, 1000)
    for elements, width, shape in test_set:
        for i in test_data:
            assert (_lt_1d(i, elements, width, shape) == laygo2.object.grid._AbsToPhyGridConverter._phy2abs_operator(i, elements, width, shape, "<")), str(
                i) + str(elements) + str(width)
            assert (_le_1d(i, elements, width, shape) == laygo2.object.grid._AbsToPhyGridConverter._phy2abs_operator(i, elements, width, shape, "<=")), str(
                i) + str(elements) + str(width)
            assert (_gt_1d(i, elements, width, shape) == laygo2.object.grid._AbsToPhyGridConverter._phy2abs_operator(i, elements, width, shape, ">")), str(
                i) + str(elements) + str(width)
            assert (_ge_1d(i, elements, width, shape) == laygo2.object.grid._AbsToPhyGridConverter._phy2abs_operator(i, elements, width, shape, ">=")), str(
                i) + str(elements) + str(width)

    test_set = [[[0], 180, 1]]
    for elements, width, shape in test_set:
        for i in test_data:
            # assert( _lt_1d(i, elements, width, shape ) ==_phy2abs_operator(i,elements, width, shape, "<" ) ) , str(i) + str(elements) + str(width)
            assert (_le_1d(i, elements, width, shape) == laygo2.object.grid._AbsToPhyGridConverter._phy2abs_operator(i, elements, width, shape, "<=")), str(
                i) + str(elements) + str(width)
            # assert (_gt_1d(i, elements, width, shape) == _phy2abs_operator(i, elements, width, shape, ">")) , str(i) + str(elements) + str(width)
            assert (_ge_1d(i, elements, width, shape) == laygo2.object.grid._AbsToPhyGridConverter._phy2abs_operator(i, elements, width, shape, ">=")), str(
                i) + str(elements) + str(width)

def test_grid_Grid():
    """
    DUT     : laygo2.object.grid.Grid
    Comment : test Grid Class
    Case    :  <, >, <=, >=, property
    """

    print("Testing 2d grid")
    g1_o  = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 100], elements=[0, 10, 20, 40, 50])
    g1_no = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 200], elements=[10, 20, 25])
    g2    = laygo2.object.grid.Grid(name='mygrid', vgrid=g1_o, hgrid=g1_no)

    test_set_index = [
        [g2[2, 5], np.array([20, 225])],
        [g2[2, :], np.array([[20, 10], [20, 20], [20, 25]])],
        [g2[0:8, 1:3], np.array([[[0, 20], [0, 25]], [[10, 20], [10, 25]], [[20, 20], [20, 25]], [[40, 20], [40, 25]],
                                [[50, 20], [50, 25]], [[100, 20], [100, 25]], [[110, 20], [110, 25]],
                                [[120, 20], [120, 25]]])],

        [g2 == [20, 10], np.array([2, 0])],
        [g2 == [[20, 10], [40, 25]], np.array([[2, 0], [3, 2]])],
        [g2.phy2abs[[20, 40], [10, 25]], np.array([[[2, 0], [2, 2]], [[3, 0], [3, 2]]])]
    ]

    test_set_op = [
        [ g2 <  [ 20,10 ],  np.array( [1, -1 ] )  ],
        [ g2 <  [ 25, 15],  np.array( [2, 0  ] )  ],
        [ g2  <  [125, 215], np.array( [7, 3])],

        [g2 <= [20, 10], np.array([2, 0])],
        [g2 <= [25, 15], np.array([2, 0])],
        [g2 <= [125, 215], np.array([7, 3])],

        [g2 > [20, 10], np.array([3, 1])],
        [g2 > [25, 15], np.array([3, 1])],
        [g2 > [125, 215], np.array([8, 4])],

        [g2 >= [20, 10], np.array([2, 0])],
        [g2 >= [25, 15], np.array([3, 1])],
        [g2 >= [125, 215], np.array([8, 4])],
    ]

    test_set_listOp = [
        [ g2 <  [ [20,10], [25,15], [125,215] ] , [ [1,-1], [2,0], [7,3] ] ],
        [g2  <= [[20, 10], [25, 15], [125, 215]], [[2, 0], [2, 0], [7, 3]]],
        [g2  > [[20, 10], [25, 15], [125, 215]], [[3, 1], [3, 1], [8, 4]]],
        [g2 >= [[20, 10], [25, 15], [125, 215]], [[2, 0], [3, 1], [8, 4]]],
    ]

    test_set_property=[
        [ g2.elements, [ [0,10,20,40,50], [10,20,25] ] ],
        [ g2.width, 100],
        [ g2.height, 200]

    ]
    test_case_map = test_set_index + test_set_op + test_set_listOp

    for q, a  in test_case_map:
        assert  np.all( q == a)
    for q, a in test_set_property:
        assert q, a

def test_grid_RoutingGrid():
    """
    DUT     : laygo2.object.grid.RoutingGrid(Grid)
    Comment : TBD
    Case    : TBD
    """

    from laygo2.object.template import NativeInstanceTemplate
    g1a = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 100], elements=[0, 10, 20, 40, 50])
    g1b = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 200], elements=[10, 20, 25])

    vwidth     = laygo2.object.grid.CircularMapping(elements=[2, 2, 2, 5, 5])
    hwidth     = laygo2.object.grid.CircularMapping(elements=[4, 4, 4])
    vextension = laygo2.object.grid.CircularMapping(elements=[1, 1, 1, 1, 1])
    hextension = laygo2.object.grid.CircularMapping(elements=[1, 1, 1])
    vlayer     = laygo2.object.grid.CircularMapping(elements=[['M1', 'drawing']] * 5, dtype=object)
    hlayer     = laygo2.object.grid.CircularMapping(elements=[['M2', 'drawing']] * 3, dtype=object)
    pvlayer    = laygo2.object.grid.CircularMapping(elements=[['M1', 'drawing']] * 5, dtype=object)
    phlayer    = laygo2.object.grid.CircularMapping(elements=[['M2', 'drawing']] * 3, dtype=object)
    tvia0  = NativeInstanceTemplate( libname='mylib', cellname='mynattemplate', bbox=[[0, 0], [100, 100]] )
    viamap = laygo2.object.grid.CircularMappingArray(elements=np.array([ [tvia0] * 5] * 3), dtype=object)
    # ivia0 = tvia0.generate(name='IV0', xy=[0, 0], transform='R0')
    # via_inst = ivia0
    # viamap = CircularMapping(elements=np.array([[via_inst]*5]*3), dtype=object)
    # g3 = RoutingGrid(name='mygrid', x=g1a, y=g1b, widthmap=width, layermap=layer, viamap=viamap)
    g3 = laygo2.object.grid.RoutingGrid(name='mygrid', vgrid=g1a, hgrid=g1b, vwidth=vwidth, hwidth=hwidth,
                                        vextension=vextension,
                                        hextension=hextension, vlayer=vlayer, hlayer=hlayer, viamap=viamap,
                                        pin_vlayer=vlayer, pin_hlayer=hlayer,
                                        xcolor=["Not MPT"], ycolor=["Not MPT"]
                                        )
    test_pair = [
        [g3[2, 5], np.array([20, 225])],
        [g3[2, :], np.array([[20, 10], [20, 20], [20, 25]])],
        [g3[0:8, 1:3], np.array([[[0, 20], [0, 25]], [[10, 20], [10, 25]], [[20, 20], [20, 25]], [[40, 20], [40, 25]],
                                 [[50, 20], [50, 25]], [[100, 20], [100, 25]], [[110, 20], [110, 25]],
                                 [[120, 20], [120, 25]]])],
        [g3.mn[40, 25], np.array([3, 2])],
        [g3 == [20, 10], np.array([2, 0])],
        [g3 == [[20, 10], [40, 25]], np.array([[2, 0], [3, 2]])],
        [g3.phy2abs[[20, 40], [10, 25]], np.array([[[2, 0], [2, 2]], [[3, 0], [3, 2]]])],
    ]


    for q, a in test_pair:
    #    #print(q, np.all(q == a))
        assert (np.all(q == a))

def test_grid_route_via_track():
    """
    DUT     : laygo2.object.grid.route_via_track()
    Comment : test RoutingGrid.rount_via_track()
    Case    :  Center-track, Upper-track, Lower-track, Overlap-track, disorder-track
    """

    tpmos_name, tnmos_name = 'pmos', 'nmos'
    pg_name = 'placement_cmos'
    libname = 'laygo2_test_1'
    templates = tech.load_templates()
    tpmos, tnmos = templates[tpmos_name], templates[tnmos_name]

    grids = tech.load_grids(templates=templates)
    pg, r12, r23 = grids['placement_cmos'], grids['routing_12_cmos'], grids['routing_23_cmos']

    lib = laygo2.object.database.Library(name=libname)
    dsn = laygo2.object.database.Design(name="test", libname=libname)
    lib.append(dsn)

    def _compare(dsn, grid, mn, track, route_golden, via_golden):
        route_golden = np.asarray(route_golden)
        via_golden = np.asarray(via_golden)

        dsn.route_via_track(grid=grid, track=track, mn=mn)
        route, via = [], []

        for name, i in dsn.rects.items():
            route.append(grid.mn(i.bbox))

        for name, i in dsn.virtual_instances.items():
            via.append(grid.mn(i.bbox))

        if np.array_equal(route_golden, np.asarray(route)) & np.array_equal(via_golden, np.asarray(via)) != True:
            print("log")
            print(route_golden)
            print(route)
            print(via_golden)
            print(via)

        return (np.array_equal(route_golden, np.asarray(route)) & np.array_equal(via_golden, np.asarray(via)))

    _mn, _track, _route_g, _via_g = [], [], [], []  ## input, input, expected, expected

    _mn.append([[0, 0], [0, 2], [2, 2]])  ## 1. Center track
    _track.append([None, 1])
    _route_g.append([bb(0, 0, 0, 1), bb(0, 1, 0, 2), bb(2, 1, 2, 2), bb(0, 1, 2, 1)])
    _via_g.append([bb(0, 1, 0, 1), bb(0, 1, 0, 1), bb(2, 1, 2, 1)])

    _mn.append([[0, 0], [0, 2], [2, 2]])  ## 2. Upper Track
    _track.append([None, 4])
    _route_g.append([bb(0, 0, 0, 4), bb(0, 2, 0, 4), bb(2, 2, 2, 4), bb(0, 4, 2, 4)])
    _via_g.append([bb(0, 4, 0, 4), bb(0, 4, 0, 4), bb(2, 4, 2, 4)])

    _mn.append([[0, 0], [0, 2], [2, 2]])  ## 3. Lower Track
    _track.append([None, -2])
    _route_g.append([bb(0, -2, 0, 0), bb(0, -2, 0, 2), bb(2, -2, 2, 2), bb(0, -2, 2, -2)])
    _via_g.append([bb(0, -2, 0, -2), bb(0, -2, 0, -2), bb(2, -2, 2, -2)])

    _mn.append([[0, 0], [0, 2], [2, 2]])  ## 4. overlap Track
    _track.append([None, 0])
    _route_g.append([bb(0, 0, 0, 2), bb(2, 0, 2, 2), bb(0, 0, 2, 0)])
    _via_g.append([bb(0, 0, 0, 0), bb(0, 0, 0, 0), bb(2, 0, 2, 0)])

    _mn.append([[2, 2], [0, 0], [0, 2]])  ## 5. sequence Track
    _track.append([None, 1])
    _route_g.append([bb(2, 1, 2, 2), bb(0, 0, 0, 1), bb(0, 1, 0, 2), bb(0, 1, 2, 1)])
    _via_g.append([bb(2, 1, 2, 1), bb(0, 1, 0, 1), bb(0, 1, 0, 1)])

    for i in range(len(_mn)):
        assert (_compare(dsn, r23, _mn[i], _track[i], _route_g[i], _via_g[i]))
        dsn.rects.clear()
        dsn.virtual_instances.clear()

def test_database_design_bbox():
    templates = tech.load_templates()
    grids     = tech.load_grids(templates)
    pg        = grids["placement_basic"]

    dsn   = laygo2.object.database.Design(name="test")
    inst0 = laygo2.object.physical.Instance(name="I0", xy=[0, 0], libname="mylib", cellname="mycell", unit_size=[100, 100], pins=None,
                                            transform='R0')
    inst1 = laygo2.object.physical.Instance(name="I1", xy=[0, 0], libname="mylib", cellname="mycell",
                                            unit_size=[100, 100], pins=None,
                                            transform='R0')
    dsn.place(grid=pg, inst= inst0, mn=[0,  0])
    dsn.place(grid=pg, inst= inst1, mn=[10, 0])

    print(dsn.bbox)
    assert( np.array_equal( dsn.bbox, [[0,0],[200,100]]))

def test_template_user():
    vinst0_pins = dict()
    pin0 = laygo2.object.physical.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
    pin1 = laygo2.object.physical.Pin(xy=[[20, 20], [30, 30]], layer=['M1', 'drawing'], netname='in')
    pin2 = laygo2.object.physical.Pin(xy=[[40, 40], [50, 50]], layer=['M1', 'drawing'], netname='in')
    vinst0_pins['in'] = laygo2.object.physical.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in', elements= [pin0,pin1,pin2])

    vinst0_native_elements = dict()
    vinst0_native_elements['R0'] = laygo2.object.physical.Rect(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'])
    #vinst0_native_elements['R1'] = laygo2.object.physical.Rect(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'])
    #vinst0_native_elements['R2'] = laygo2.object.physical.Rect(xy=[[0, 0], [100, 100]], layer=['prBoundary', 'drawing'])
    print("Create V Instance")
    vinst0 = laygo2.object.physical.VirtualInstance(name='I0', libname='mylib', cellname='myvcell', xy=[100, 100],
                                                    native_elements=vinst0_native_elements, unit_size=[100, 100], pins=vinst0_pins, transform='MX')

    print("END V Instance")
    print(vinst0.xy)
    vinst1 = laygo2.object.physical.VirtualInstance(name='I1', libname='mylib', cellname='myvcell', xy=[100, 100],
                                                    native_elements=vinst0_native_elements, unit_size=[100, 100],
                                                    pins=vinst0_pins, transform='R0')

    vinst0.xy=[500,500]
    print("compare")
    for i in range(3):
#        print(vinst1.pins["in"][i].xy)
        print(vinst0.pins["in"][i].xy)
    pass
"""
def test_database_design_test():

    g1_y     = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 50], elements=[0, 10, 20, 30, 40])
    g1_x_cut = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 5], elements=[0])
    g1_x     = laygo2.object.grid.OneDimGrid(name='mygrid', scope=[0, 10], elements=[0])

    templates = tech.load_templates()
    grids     = tech.load_grids(templates)
    pg        = grids["placement_cmos"]
    viamap    = 5*["via_r23_default"]
    vmap_mapped =[]
    for via_name in viamap:
        vmap_mapped.append( templates[ via_name ] )
    vmap_mapped = np.asarray([vmap_mapped])
    vmap_mapped2 = laygo2.object.grid.CircularMappingArray(elements=vmap_mapped, dtype=object)

    grid_cut = laygo2.object.grid.RoutingGrid(name='test', vgrid=g1_x_cut, hgrid=g1_y, viamap=vmap_mapped2
                                              ,vwidth=np.zeros(4), hwidth=[0],
                                              vlayer=None, hlayer=None, pin_vlayer=None, pin_hlayer=None,
                                              vextension=None, hextension=None )

    grid     = laygo2.object.grid.RoutingGrid(name='test', vgrid=g1_x, hgrid=g1_y, viamap=vmap_mapped2
                                              ,vwidth=np.zeros(4), hwidth=[0],
                                              vlayer=None, hlayer=None, pin_vlayer=None, pin_hlayer=None,
                                              vextension=None, hextension=None )

    dsn = laygo2.object.database.Design_test(name="test")
    r     = grid
    r_cut = grid_cut

    bbox_top  = np.array( [ [0,0], [300,290] ] ) ## instance bbox union


    rect    = [0]*7
    rect_er = [0]*4
    rect[0] = laygo2.object.physical.Rect(xy=[ [0, 10], [30, 10]],   layer=['M0', 'drawing'], netname='net0', hextension = 5, vextension = 5)
    rect[1] = laygo2.object.physical.Rect(xy=[ [0, 10], [30, 10]],   layer=['M0', 'pin'],     netname='net0', hextension = 5, vextension = 5)

    # center space error
    rect[2]    = laygo2.object.physical.Rect(xy=[[60,  40],  [90, 40]], layer=['M0', 'drawing'], netname='net0',hextension = 5, vextension = 5)
    rect_er[0] = laygo2.object.physical.Rect(xy=[[120, 40],  [150, 40]], layer=['M0', 'drawing'], netname='er0',hextension = 5, vextension = 5)
    rect[3]    = laygo2.object.physical.Rect(xy=[[210,  40], [240, 40]], layer=['M0', 'drawing'], netname='net0',hextension = 5, vextension = 5)

    # center & edge space error
    rect[4]    = laygo2.object.physical.Rect(xy=[[60, 60], [90, 60]], layer=['M0', 'drawing'], netname='net0',hextension = 5, vextension = 5)
    rect_er[1] = laygo2.object.physical.Rect(xy=[[120, 60], [150, 60]], layer=['M0', 'drawing'], netname='er1',hextension = 5, vextension = 5 )
    rect_er[2] = laygo2.object.physical.Rect(xy=[[180, 60], [270, 60]], layer=['M0', 'drawing'], netname='er2',hextension = 5, vextension = 5)

    # edge with pin & overlap
    rect[5]    = laygo2.object.physical.Rect(xy=[[210, 90], [240, 90]], layer=['M0', 'drawing'], netname='net0',hextension = 5, vextension = 5)
    rect[6] = laygo2.object.physical.Rect(xy=[[210, 90], [270, 90]], layer=['M0', 'pin'], netname='net0',hextension = 5, vextension = 5)



    vinst0_native_elements = dict()
    vinst0_native_elements['R0'] = laygo2.object.physical.Rect(xy=[[ 200, 100], [220, 100]], layer=['M0', 'drawing'])

    vinstr0 = laygo2.object.physical.VirtualInstance(name='I0', libname='mylib', cellname='myvcell', xy=[0, 0],
                                                     native_elements=vinst0_native_elements,
                                                     unit_size=[300, 300], pins=None, transform='R0')

    vinstmx = laygo2.object.physical.VirtualInstance(name='I1', libname='mylib', cellname='myvcell', xy=[0, 0],
                                                     native_elements=vinst0_native_elements,
                                                     unit_size=[300, 300], pins=None, transform='MX')

    rect_er[3] = laygo2.object.physical.Rect(xy=[[100, 100], [180, 100]], layer=['M0', 'drawing'], netname='er2',
                                             hextension=5, vextension=5)

    dsn.place(grid=pg, inst=[ vinstr0, vinstmx] , mn=[0, 0])

    for i, obj in enumerate( rect + rect_er):
        dsn.append(obj)

    print("start!")

    dsn.rect_space("M0", xy = bbox_top, grid=r, grid_cut = r_cut, space_min = 50 )
    result=[]

    golden=[[105,40],[105,60],[165,60],[190,100],[300,60]]
    for name, obj in dsn.virtual_instances.items():
        if "contact" in obj.native_elements:
            result.append(obj.xy)

    result=np.unique( np.array(result), axis=0 )

    assert(np.array_equal(result, golden))
"""
def test_physical_virtualInstace_get_element_position():
    dsn = laygo2.object.database.Design(name="test")
    templates = tech.load_templates()
    grids = tech.load_grids(templates=templates)
    pg = grids["placement_cmos"]
    vinst0_pins = dict()
    vinst0_pins['in'] = laygo2.object.physical.Pin(xy=[[0, 100], [100, 100]], layer=['M1', 'drawing'], netname='in')
    vinst0_native_elements = dict()
    vinst0_native_elements['R0'] = laygo2.object.physical.Rect(xy=[[ 50, 100], [400 , 100]], layer=['M1', 'drawing'])
    vinstr0 = laygo2.object.physical.VirtualInstance(name='I0', libname='mylib', cellname='myvcell', xy=[0, 0],
                                                    native_elements=vinst0_native_elements,
                                                    unit_size=[500, 500], pins=vinst0_pins, transform='R0')
    vinstmx = laygo2.object.physical.VirtualInstance(name='I1', libname='mylib', cellname='myvcell', xy=[0, 0],
                                                    native_elements=vinst0_native_elements,
                                                    unit_size=[500, 500], pins=vinst0_pins, transform='MX')
    vinstmy = laygo2.object.physical.VirtualInstance(name='I2', libname='mylib', cellname='myvcell', xy=[0, 0],
                                                    native_elements=vinst0_native_elements,
                                                    unit_size=[500, 500], pins=vinst0_pins, transform='MY')
    vinstr90 = laygo2.object.physical.VirtualInstance(name='I3', libname='mylib', cellname='myvcell', xy=[0, 0],
                                                    native_elements=vinst0_native_elements,
                                                    unit_size=[500, 500], pins=vinst0_pins, transform='R90')

    insts = [vinstr0, vinstmx, vinstmy, vinstr90 ]
    dsn.place(grid=pg, inst= insts, mn=[0,0] )

    print("start")
    for inst in insts:
        print(inst.xy, inst.transform)
    setr0 = np.array( [ [50,  100], [400, 100 ] ])
    setmx= np.array(  [ [550,  400], [900, 400 ] ]) # y
    setmy = np.array( [ [1100, 100], [1450, 100 ] ])
    setr90 = np.array( [[1600, 400], [1950, 400 ] ]) # y


    assert( np.array_equal( setr0, vinstr0.get_element_position( vinstr0.native_elements["R0"] ) ) )
    assert (np.array_equal( setmx, vinstmx.get_element_position(vinstmx.native_elements["R0"])))
    assert (np.array_equal( setmy, vinstmy.get_element_position(vinstmy.native_elements["R0"])))
    #assert (np.array_equal( setr90, dsn.get_xy_virtualInstance(vinstr90, vinstr90.native_elements["R0"])))









#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
The physical module implements classes for various physical layout objects.
The types of objects supported are summarized below:

PhysicalObject - a base class for physical layout objects.

IterablePhysicalObject - a base class for iterable physical objects (eg. arrayed instances).

PhysicalObjectGroup - defines a group of physical objects.

Rect - defines a rect.

Path - defines a path.

Pin - defines a pin.

Text - defines a text label.

Instance - defines an instance.

VirtualInstance - defines a virtual instance composed of multiple physical objects.

"""

__author__ = "Jaeduk Han"
__maintainer__ = "Jaeduk Han"
__status__ = "Prototype"

import numpy as np
# from copy import deepcopy
import laygo2.util.transform as tf

class PhysicalObject:
    """
    The base class of physical layout objects, which has physical coordinate information.

    Attributes
    ----------
    name : str
    xy : numpy.ndarray
    bbox : numpy.ndarray
    master : PhysicalObject or None
    params : dict
    pointers : dict
    left : numpy.ndarray
    right : numpy.ndarray
    top : numpy.ndarray
    bottom : numpy.ndarray
    center : numpy.ndarray
    bottom_left : numpy.ndarray
    bottom_right : numpy.ndarray
    top_left : numpy.ndarray
    top_right : numpy.ndarray

    Methods
    -------
    __init__()

    Notes
    -----
    Reference in Korean:
    물리 객체들의 기본 클래스, 물리적 좌표정보를 갖고 있다.  
    """

    def _get_xy(self):
        """numpy.ndarray(dtype=numpy.int): Get the x and y coordinate values of this object."""
        return self._xy

    def _set_xy(self, value):
        """numpy.ndarray(dtype=numpy.int): Set the x and y coordinate values of this object."""
        self._xy = np.asarray(value, dtype=np.int)
        self._update_pointers()

    name = None
    """attribute
    str: Physical object name.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.name 
    “test”

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_name.png

    Reference in Korean:
    str: Physical 객체의 이름.
    """

    _xy = np.zeros(2, dtype=np.int)
    """numpy.ndarray(dtype=numpy.int): The internal variable of xy."""

    xy = property(_get_xy, _set_xy)
    """attribute
    numpy.ndarray: Physical object coordinates.
        xy coordinates.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.xy 
    [ [0,0], [200,200] ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_xy.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 좌표.
    """

    master = None
    """attribute
    numpy.ndarray: Physical object master.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.master
    None

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_master.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 master.
    """

    params = None
    """attribute
    dict: Physical object attributes.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params={maxI=0.005}) 
    >>> physical.params 
    {‘maxI’: 0.005 }

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_params.png

    Reference in Korean:
    dict: Physical 객체의 속성.
    """

    pointers = None
    """attribute
    dict: Major coordinates of a physical object.
        key: direction, value: Coordinate value

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.pointers 
    {'left': array([0, 100]), 'right': array([200, 100]),
     'bottom': array([100, 0]), 'top': array([100, 200]),
     'bottom_left': array([0, 0]), 'bottom_right': array([200, 0]),
     'top_left': array([0, 200]), 'top_right': array([200, 200]),
     ‘center’: array( [100, 100] ) 
    }

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_pointers.png

    Reference in Korean:
    dict: Physical 객체의 주요좌표들.
        key: 방향, value: 좌표값
    """

    # frequently used pointers
    left = None
    """attribute
    numpy.ndarray: Left pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.left
    [ 0, 100 ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_left.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 left pointer.
    """
    right = None
    """attribute
    numpy.ndarray: Right pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.right 
    [ 200, 100 ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_right.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 right pointer.
    """ 
    top = None
    """attribute
    numpy.ndarray: Top pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.top 
    [ 100, 200 ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_top.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 top pointer.
    """
    bottom = None
    """attribute
    numpy.ndarray: Bottom pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.bottom 
    [ 100, 0 ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_bottom.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 bottom pointer.
    """ 
    center = None
    """attribute
    numpy.ndarray: Center pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.center 
    [ 100, 100 ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_center.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 center pointer.
    """ 
    bottom_left = None 
    """attribute
    numpy.ndarray: Bottom_left pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.bottom_left 
    [ 0,0 ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_bottom_left.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 bottom_left pointer.
    """
    bottom_right = None
    """attribute
    numpy.ndarray: Bottom_right pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.bottom_right 
    [200, 0]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_bottom_right.png 

    Reference in Korean:
    numpy.ndarray: Physical 객체의 bottom_right pointer.
    """ 
    top_left = None
    """attribute
    numpy.ndarray: Top_left pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.top_left 
    [ 0, 200 ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_top_left.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 top_left pointer.
    """
    top_right = None
    """attribute
    numpy.ndarray: Top_right pointer of a physical object.

    

    Examples
    --------
    >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
    >>> physical.top_right 
    [200, 200]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_top_right.png

    Reference in Korean:
    numpy.ndarray: Physical 객체의 top_right pointer.
    """

    @property
    def bbox(self):
        """attribute
        numpy.ndarray: Physical object bbox.

        

        Examples
        --------
        >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params=None) 
        >>> physical.xy 
        [ [0,0], [200,200] ]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_PhysicalObject_bbox.png

        Reference in Korean:
        numpy.ndarray: Physical 객체의 bbox.
        """
        return np.sort(np.array([self.xy[0, :], self.xy[1, :]]), axis=0)

    def __init__(self, xy, name=None, params=None):
        """
        Constructor function of PhysicalObject class.

        Parameters
        ----------
        xy : numpy.ndarray
            physical coordinates of object, bbox.
        name : str
            object name.
        params : dict
            dictionary having attributes.

        Returns
        -------
        laygo2.PhysicalObject

        

        Examples
        --------
        >>> physical = object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params={'maxI’: 0.005}) 
        >>> print(physical) 
        <laygo2.object.physical.PhysicalObject object at 0x000001ECF0022948>
         name: test,
         class: PhysicalObject,
         xy: [[0, 0], [200, 200]],
         params: {'maxI': 0.005}
        
        Notes
        -----
        Reference in Korean:
        PhysicalObject 클래스의 생성자 함수.
        파라미터
        xy(numpy.ndarray): 객체의 물리적 좌표, bbox
        name(str): 객체의 이름
        params(dict): 속성을 갖고 있는 Dictionary
        반환값
        laygo2.PhysicalObject
        참조
        없음
        """

        self.name = name
        # Initialize pointers.
        self.pointers = dict()
        self.pointers['left'] = np.array([0, 0], dtype=np.int)
        self.pointers['right'] = np.array([0, 0], dtype=np.int)
        self.pointers['bottom'] = np.array([0, 0], dtype=np.int)
        self.pointers['top'] = np.array([0, 0], dtype=np.int)
        self.pointers['bottom_left'] = np.array([0, 0], dtype=np.int)
        self.pointers['bottom_right'] = np.array([0, 0], dtype=np.int)
        self.pointers['top_left'] = np.array([0, 0], dtype=np.int)
        self.pointers['top_right'] = np.array([0, 0], dtype=np.int)
        self.left = self.pointers['left']
        self.right = self.pointers['right']
        self.bottom = self.pointers['bottom']
        self.top = self.pointers['top']
        self.bottom_left = self.pointers['bottom_left']
        self.bottom_right = self.pointers['bottom_right']
        self.top_left = self.pointers['top_left']
        self.top_right = self.pointers['top_right']

        self.params = params  # deepcopy(params)  # deepcopy increases the memory usage.
        self.xy = xy

    def __str__(self):
        """Return the summary of this object."""
        return self.summarize()

    def summarize(self):
        """Return the summary of this object."""
        name = 'None' if self.name is None else self.name
        return \
            self.__repr__() + " " \
            "name: "   + name + ", " + \
            "class: "  + self.__class__.__name__ + ", " + \
            "xy: "     + str(self.xy.tolist()) + ", " + \
            "params: " + str(self.params) + ", " + \
            ""

    def _update_pointers(self):
        """Update pointers of this object. Called when the xy-coordinate of this object is updated."""
        xy_left = np.diag(np.dot(np.array([[1, 0], [0.5, 0.5]]), self.bbox)).astype(np.int)
        xy_right = np.diag(np.dot(np.array([[0, 1], [0.5, 0.5]]), self.bbox)).astype(np.int)
        xy_bottom = np.diag(np.dot(np.array([[0.5, 0.5], [1, 0]]), self.bbox)).astype(np.int)
        xy_top = np.diag(np.dot(np.array([[0.5, 0.5], [0, 1]]), self.bbox)).astype(np.int)
        xy_bottom_left = np.diag(np.dot(np.array([[1, 0], [1, 0]]), self.bbox)).astype(np.int)
        xy_bottom_right = np.diag(np.dot(np.array([[0, 1], [1, 0]]), self.bbox)).astype(np.int)
        xy_top_left = np.diag(np.dot(np.array([[1, 0], [0, 1]]), self.bbox)).astype(np.int)
        xy_top_right = np.diag(np.dot(np.array([[0, 1], [0, 1]]), self.bbox)).astype(np.int)
        xy_center    = np.diag(np.dot(np.array([[0.5, 0.5], [0.5, 0.5]]), self.bbox)).astype(np.int)
        self.pointers['left'] = xy_left
        self.pointers['right'] = xy_right
        self.pointers['bottom'] = xy_bottom
        self.pointers['top'] = xy_top
        self.pointers['bottom_left'] = xy_bottom_left
        self.pointers['bottom_right'] = xy_bottom_right
        self.pointers['top_left'] = xy_top_left
        self.pointers['top_right'] = xy_top_right
        self.pointers['center'] = xy_center
        self.left = self.pointers['left']
        self.right = self.pointers['right']
        self.bottom = self.pointers['bottom']
        self.top = self.pointers['top']
        self.bottom_left = self.pointers['bottom_left']
        self.bottom_right = self.pointers['bottom_right']
        self.top_left = self.pointers['top_left']
        self.top_right = self.pointers['top_right']
        self.center = self.pointers['center']


class IterablePhysicalObject(PhysicalObject):
    """
    The base class for iterable objects among physics objects.
    Basic class of entities capable of iterable operation among physical entities.

    Attributes
    ----------
    name : str
    xy : numpy.ndarray
    bbox : numpy.ndarray
    master : PhysicalObject or None
    params : dict
    pointers : dict
    left : numpy.ndarray
    right : numpy.ndarray
    top : numpy.ndarray
    bottom : numpy.ndarray
    center : numpy.ndarray
    bottom_left : numpy.ndarray
    bottom_right : numpy.ndarray
    top_left : numpy.ndarray
    top_right : numpy.ndarray
    elements : numpy.ndarray
    shape : numpy.ndarray

    Methods
    -------
    __init__() : laygo2.IterablePhysicalObject
    ndenumerate() : numpy.ndenumerate

    Reference in Korean:
    물리 객체들 중 iterable 연산이 가능한 객체들의 기본 클래스.
    """

    elements = None
    """attribute
    numpy.ndarray: The lower member of objects.

    

    Examples
    --------
    >>> phy0 = physical.IterablePhysicalObject( xy=[[0, 0], [100, 100]], name="test" ) 
    >>> phy1 = physical.IterablePhysicalObject( xy=[[0, 0], [200, 200]], name="test") 
    >>> phy2 = physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test") 
    >>> element = [phy0, phy1, phy2] 
    >>> iphy0 = physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test", elements = element)
    >>> iphy0.elements 
    [physical.IterablePhysicalObject object,
     physical.IterablePhysicalObject object,
     physical.IterablePhysicalObject object ]

    Notes
    -----
    Reference in Korean:
    numpy.ndarray: 객체의 하위 구성원.
    """
    def _get_xy(self):
        """numpy.ndarray(dtype=numpy.int): Get the x and y coordinate values of this object."""
        return self._xy

    def _set_xy(self, value):
        """numpy.ndarray(dtype=numpy.int): Set the x and y coordinate values of this object."""
        # Update the coordinate value of its elements.
        self._update_elements(xy_ofst=value - self.xy)
        # Update the coordinate value of the object itself.
        PhysicalObject._set_xy(self, value=value)

    xy = property(_get_xy, _set_xy)

    @property
    def shape(self):
        """attribute
        numpy.ndarray: Shape of object element.

        

        Examples
        --------
        >>> phy0 = physical.IterablePhysicalObject( xy=[[0, 0], [100, 100]], name="test" ) 
        >>> phy1 = physical.IterablePhysicalObject( xy=[[0, 0], [200, 200]], name="test") 
        >>> phy2 = physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test") 
        >>> element = [phy0, phy1, phy2] 
        >>> iphy0 = physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test", elements = element)
        >>> iphy0.shape 
        [3]

        Notes
        -----
        Reference in Korean:
        numpy.ndarray: 객체의 element의 shape.
        """
        if self.elements is None:
            return None
        else:
            return np.array(self.elements.shape, dtype=np.int)

    def __init__(self, xy, name=None, params=None, elements=None):
        """
        Constructor function of IterablePhysicalObject class.

        Parameters
        ----------
        xy : numpy.ndarray
            physical coordinates of object, bbox.
        name : str
            object name.
        params : dict
            dictionary having attributes.
        elements : dict
            dictionary having unit object.
        
        Returns
        -------
        laygo2.IterablePhysicalObject

        See Also
        --------
        Class PhysicalObject

        Examples
        --------
        >>> phy0 = physical.IterablePhysicalObject( xy=[[0, 0], [100, 100]], name="test" ) 
        >>> phy1 = physical.IterablePhysicalObject( xy=[[0, 0], [200, 200]], name="test") 
        >>> phy2 = physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test") 
        >>> element = [phy0, phy1, phy2] 
        >>> iphy0 = physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test", elements = element)
        >>> print(iphy0) 
        <laygo2.object.physical.IterablePhysicalObject object at 0x0000013219648F08> 
        name: test, class: IterablePhysicalObject, 
        xy: [[0, 0], [300, 300]], params: None

        Notes
        -----
        Reference in Korean:
        IterablePhysicalObject 클래스의 생성자 함수.
        파라미터
        xy(numpy.ndarray): 객체의 물리적 좌표, bbox
        name(str): 객체의 이름
        params(dict): 속성을 갖는 Dictionary
        elements(dict): 단위 객체를 갖는 Dictionary
        반환값
        laygo2.IterablePhysicalObject
        참조
        Class PhysicalObject
        """
        PhysicalObject.__init__(self, xy=xy, name=name, params=params)
        if elements is None:
            self.elements = None
        else:
            self.elements = np.asarray(elements)

    def __getitem__(self, pos):
        """Return the sub-elements of this object, based on the pos parameter."""
        return self.elements[pos]

    def __setitem__(self, pos, item):
        """Set the sub-elements of this object, based on the pos and item parameter. """
        self.elements[pos] = item

    def __iter__(self):
        """Iterator function. Directly mapped to the iterator of the elements attribute of this object."""
        return self.elements.__iter__()

    def __next__(self):
        """Iterator function. Directly mapped to the iterator of the elements attribute of this object."""
        return self.elements.__next__()

    def ndenumerate(self):
        """Enumerate over the element array. Calls np.ndenumerate() of the elements of this object."""
        return np.ndenumerate(self.elements)

    def _update_elements(self, xy_ofst):
        """Update xy-coordinates of this object's elements. An internal function for _set_xy()"""
        #print("aa?")
        if np.all(self.elements is not None):
            # Update the x and y coordinate values of elements.
            for n, e in self.ndenumerate():
                if e is not None:
                    e.xy = e.xy + xy_ofst

class PhysicalObjectGroup(IterablePhysicalObject):
    """
    A grouped physical object. Intended to be generated as a group in Virtuoso (not implemented yet).
    """
    # TODO: implement this.

    def summarize(self):
        """Return the summary of the object information."""
        return IterablePhysicalObject.summarize(self) + "\n" + \
               "  elements: " + str(self.elements)

    def __init__(self, xy, name=None, params=None, elements=None):
        """
        Constructor.

        Parameters
        ----------
        xy : numpy.ndarray(dtype=int)
            The coordinate of this object represented as a Numpy array [x, y].
        name : str, optional
            The name of the object.
        params : dict or None
            The dictionary that contains the parameters of this object, with parameter names as keys.
        elements : numpy.ndarray(dtype=LayoutObject) or None
            The iterable elements of the object.
        """
        IterablePhysicalObject.__init__(self, xy=xy, name=name, params=params, elements=elements)


'''
# Deprecated as PhysicalObjectGroup can be used instead in most cases.
class PhysicalObjectArray(np.ndarray):
    """LayoutObject array class for containing multiple layout objects. Subclassing ndarray to utilize advance slicing
     functions."""
    name = None
    """str: the name of the object."""

    params = None
    """dict or None: parameters of the object. """

    _xy = None  # np.zeros(2, dtype=np.int)
    """numpy.ndarray(dtype=numpy.int): the internal variable of xy."""

    def get_xy(self):
        """numpy.ndarray(dtype=numpy.int): gets the x and y coordinate values of this object."""
        return self._xy

    def set_xy(self, value):
        """numpy.ndarray(dtype=numpy.int): sets the x and y coordinate values of this object."""
        if value is None:
            self._xy = value
        else:
            self._xy = np.asarray(value, dtype=np.int)

    xy = property(get_xy, set_xy)
    """numpy.ndarray(detype=numpy.int): the x and y coordinate values of the object."""

    def moveby(self, delta):
        """move the array and child objects by delta."""
        self.xy = self.xy + delta
        for i in self:
            i.xy = i.xy + delta

    def __new__(cls, input_array, name=None, xy=None, params=None):
        """
        Constructor for ndarray subclasses - check the NumPy manual for details.

        Parameters
        ----------
        input_array : np.ndarray
            An array of LayoutObject objects.
        name : str
            The name of the array.
        xy : numpy.ndarray(dtype=int)
            The xy-coordinate of the object. The format is [x0, y0].
        params : dict
            Additional parameters of the array.
        """
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj.name = name
        obj.xy = None if xy is None else np.asarray(xy, dtype=np.int)
        obj.params = params
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        """
        Array finalizing function for subclassing ndarray - check the NumPy manual for details
        """
        if obj is None: return
        # Transfer parameters
        self.name = getattr(obj, 'name', None)
        self.xy = getattr(obj, 'xy', None)
        self.params = getattr(obj, 'params', None)

    def __str__(self):
        """Returns the summary of this object."""
        return self.summarize()

    def summarize(self):
        """Summarizes object information."""
        return "  " + \
               "name:" + self.name + ", " + \
               "class:" + self.__class__.__name__ + ", " + \
               "shape:" + str(self.shape) + ", " + \
               "xy:" + str(self.xy) + ", " + \
               "params:" + str(self.params) + "\n" + \
               "  elements:" + str(np.ndarray.__str__(self)) + "\n"
'''

class Rect(PhysicalObject):
    """
    Rectangle object class.

    Attributes
    ----------
    layer : [str, str]
    netname : str
    hextension : int
    vextension : int
    height : int
    width : int
    color : str
    size : np.array(dtype=np.int)

    Methods
    -------
    __init__()

    Notes
    -----
    Reference in Korean:
    사각형 객체 클래스.
    """

    layer = None
    """attribute
    numpy.ndarray: Layer information of object.

    

    Examples
    --------
    >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
    >>> rect0.layer 
    [ ‘M1’, ‘drawing’ ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_layer.png

    Reference in Korean:
    numpy.ndarray: 객체의 Layer 정보.
    """
    netname = None
    """attribute
    str: Object node name.

    

    Examples
    --------
    >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
    >>> rect0.netname 
    “net0”

    Notes
    -----
    Reference in Korean:
    str: 객체의 노드명.
    """
    hextension = 0
    """attribute
    int: Extension value of object in the horizontal direction.

    

    Examples
    --------
    >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
    >>> rect0.hextension 
    20

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_hextension.png

    Reference in Korean:
    int: 객체의 수평방향으로의 확장값.
    """
    vextension = 0
    """attribute
    int: Extension value of object in the vertical direction.

    

    Examples
    --------
    >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
    >>> rect0.vextension 
    20

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_vextension.png

    Reference in Korean:
    int: 객체의 수직방향으로의 확장값.
    """
    color = None
    """attribute
    int: Object color.

    

    Examples
    --------
    >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
    >>> rect0.color 
    “notMPT”

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_color.png

    Reference in Korean:
    int: 객체의 color.
    """

    @property
    def height(self):
        """attribute
        int: Object height.

        

        Examples
        --------
        >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
        >>> rect0.height 
        100

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_height.png

        Reference in Korean:
        int: 객체의 높이.
        """
        return abs(self.xy[0, 1] - self.xy[1, 1])

    @property
    def width(self):
        """attribute
        int: Object width.

        

        Examples
        --------
        >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
        >>> rect0.width 
        100

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_width.png

        Reference in Korean:
        int: 객체의 폭.
        """
        return abs(self.xy[0, 0] - self.xy[1, 0])

    @property
    def height_vec(self):
        """numpy.ndarray(dtype=int): Return np.array([0, height])."""
        return np.array([0, self.height])

    @property
    def width_vec(self):
        """numpy.ndarray(dtype=int): Return np.array([width, 0])."""
        return np.array([self.width, 0])

    @property
    def size(self):
        """attribute
        numpy.ndarray: Object size.

        

        Examples
        --------
        >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
        >>> rect0.size 
        [100, 100]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_size.png

        Reference in Korean:
        numpy.ndarray: 객체의 크기.
        """
        return np.array([self.width, self.height])

    def __init__(self, xy, layer, color=None, hextension=0, vextension=0, name=None, netname=None, params=None):
        """
        Constructor function of Rectangle class.

        Parameters
        ----------
        xy : numpy.ndarray
            physical coordinates of object, bbox.
        layer : list
            layer information of object.
        hextension : int
            horizontal extension value of object.
        vextension : int
            vertical extension value of object.
        name : str
            object name.
        netname : str
            node name of object.
        params : dict
            dictionary having attributes.
        color : str, optional.
            object color.
        
        Returns
        -------
        laygo2.Pin

        See Also
        --------
        Class PhysicalObject

        Examples
        --------
        >>> rect0 = physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing’], netname='net0', hextension=20, vextension=20)
        >>> print( rect0 ) 
        <laygo2.object.physical.Rect object at 0x00000283542688C8> name: None, class: Rect, xy: [[0, 0], [100, 100]], params: None, , layer: ['M1', 'drawing'], netname: net0

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_rect_init.png

        Reference in Korean:
        Rect 클래스의 생성자 함수
        파라미터
        xy(numpy.ndarray): 객체의 물리적 좌표, bbox
        layer(list): 객체의 layer 정보
        hextension(int): 객체의 수평부 확장값
        vextension(int): 객체의 수직 확장값
        name(str): 객체의 이름
        netname(str): 객체의 노드 명
        params(dict): 속성을 갖는 Dictionary [optional]
        color(str): 객체의 color [optional]
        반환값
        laygo2.Pin
        참조
        Class PhysicalObject
        """
        self.layer = layer
        if netname is None:
            self.netname = name
        else:
            self.netname = netname
        self.hextension = hextension
        self.vextension = vextension
        self.color = color
        PhysicalObject.__init__(self, xy=xy, name=name, params=params)

    def align(self, rect2):
        """
        Match the length of self and argument, assuming either width or height is zero.

        Parameters
        ----------
        rect2 : Rect
            Rect which want to be matched
        """
        index = 0
        r0 = self
        r1 = rect2
        if r0.xy[0][0] == r0.xy[1][0]: # width is zero
                index = 1

        pnt = np.zeros([2, 2], dtype=int)
        pnt[0][1] = r0.bbox[1][index] # tr
        pnt[1][1] = r1.bbox[1][index] # tr
        pnt[0][0] = r0.bbox[0][index] # bl
        pnt[1][0] = r1.bbox[0][index] # bl

        if pnt[1][1] > pnt[0][1]:  # p1-top is upper then p0-top
            _xy = r0.bbox  # r0 correction
            _xy[1][index] = pnt[1][1]
            r0.xy = _xy
        elif pnt[1][1] < pnt[0][1]: #p1-top is lower then p0-top
            _xy = r1.bbox  # r1 correction
            _xy[1][index] = pnt[0][1]
            r1.xy = _xy

        if pnt[1][0] < pnt[0][0]:  # p1-bottom is lower then p0-bottom
            _xy = r0.bbox  # r0 correction
            _xy[0][index] = pnt[1][0]
            r0.xy = _xy
        elif pnt[1][0] > pnt[0][0]:
            _xy = r1.bbox  # r1 correction
            _xy[0][index] = pnt[0][0]
            r1.xy = _xy

    def summarize(self):
        """Return the summary of the object information."""
        return PhysicalObject.summarize(self) + ", " + \
                                              "layer: " + str(self.layer) + ", " + \
                                              "netname: " + str(self.netname)


class Path(PhysicalObject):
    """
    Path object class

    Attributes
    ----------
    layer : list
    netname : str
    width : int
    extension : int

    Methods
    -------
    __int__()

    Notes
    -----
    Reference in Korean:
    Path 객체 클래스.
    """
    # TODO: implement pointers.

    layer = None
    """attribute
    list: Layer information of object.

    Notes
    -----
    Reference in Korean:
    list: 객체의 레이어 정보.
    """
    netname = None
    """attribute
    str: Object node name.

    

    Examples
    --------
    >>> path0 = physical.Path(xy=[[0, 0], [0, 100]], width=10, extension=5, layer=['M1', 'drawing’], netname='net0’)
    >>> path0.netname 
    “net0”

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_path_netname.png 

    Reference in Korean:
    str: 객체의 노드명.
    """
    width = None
    """attribute
    int: Object width.

    

    Examples
    --------
    >>> path0 = physical.Path(xy=[[0, 0], [0, 100]], width=10, extension=5, layer=['M1', 'drawing’], netname='net0’)
    >>> path0.width 
    10

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_path_width.png

    Reference in Korean:
    int: 객체의 폭.
    """
    extension = 0
    """attribute
    int: Extension value of object.

    

    Examples
    --------
    >>> path0 = physical.Path(xy=[[0, 0], [0, 100]], width=10, extension=5, layer=['M1', 'drawing’], netname='net0’)
    >>> path0.extension 
    5

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_path_extension.png

    Reference in Korean:
    int: 객체의 확장값.
    """

    @property
    def bbox(self):
        return np.sort(np.array([self.xy[0], self.xy[-1]]), axis=0)

    def _update_pointers(self):
        """Update pointers of this object. Called when the value of xy of this object is updated."""
        pass

    def __init__(self, xy, layer, width, extension=0, name=None, netname=None, params=None):
        """
        Generate Path object.

        Parameters
        ----------
        xy : numpy.ndarray
            physical coordinates of object, bbox.
        layer : list
            layer information of object.
        width : int
            object width.
        extension : int
            extension value of object.
        name : str
            object name.
        netname : str
            node name of object.
        params : dict
            dictionary having attributes.

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        Class PhysicalObject

        Examples
        --------
        >>> path0 = physical.Path(xy=[[0, 0], [0, 100]], width=10, extension=5, layer=['M1', 'drawing’], netname='net0’)
        >>> print( path0 ) 
        <laygo2.object.physical.Path object at 0x00000280D1F3CE88> name: None, class: Path, xy: [[0, 0], [0, 100]], params: None, width: 10, extension: 5, layer: ['M1', 'drawing'], netname: net0

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_path_init.png

        Reference in Korean:
        Path 객체 생성
        파라미터
        xy(numpy.ndarray): 객체의 물리적 좌표, bbox
        layer(list): 객체의 layer 정보
        width(int): 객체의 폭
        extension(int): 객체의 확장값
        name(str): 객체의 이름
        netname(str): 객체의 노드 명
        params(dict): 속성을 갖는 Dictionary
        반환값
        numpy.ndarray
        참조
        Class PhysicalObject
        """
        self.layer = layer
        self.width = width
        self.extension = extension
        self.netname = netname
        PhysicalObject.__init__(self, xy=xy, name=name, params=params)
        self.pointers = dict()  # Pointers are invalid for Path objects.

    def summarize(self):
        """Return the summary of the object information."""
        return PhysicalObject.summarize(self) + ", " + \
               "width: " + str(self.width) + ", " + \
               "extension: " + str(self.extension) + ", " + \
               "layer: " + str(self.layer) + ", " + \
               "netname: " + str(self.netname)

class Pin(IterablePhysicalObject):
    """
    Pin object class.

    Attributes
    ----------
    layer : list
    netname : str
    height : int
    width : int
    size : numpy.ndarray

    Methods
    -------
    __init__()
    ndenumerate()

    Notes
    -----
    Reference in Korean:
    Pin 객체 클래스.
    """

    layer = None
    """attribute
    numpy.ndarray: Layer information of object.

    

    Examples
    --------
    >>> pin0 = physical.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'direction': 'input'})path0.width
    >>> pin0.layer 
    ['M1', 'drawing']

    Notes
    -----
    numpy.ndarray: 객체의 Layer 정보.
    """

    netname = None
    """attribute
    str: Object node name.

    

    Examples
    --------
    >>> pin0 = physical.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'direction': 'input'})path0.width
    >>> pin0.netname 
    “net0”

    Notes
    -----
    str: 객체의 노드명.
    """

    master = None
    """Instance: The master instance of the pin. Used for instance pins only."""

    @property
    def height(self):
        """attribute
        int: Object height.

        

        Examples
        --------
        >>> pin0 = physical.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'direction': 'input'})path0.width
        >>> pin0.height 
        100

        Notes
        -----
        Reference in Korean:
        int: 객체의 높이.
        """
        return abs(self.xy[0, 1] - self.xy[1, 1])

    @property
    def width(self):
        """attribute
        int: Object width.

        

        Examples
        --------
        >>> pin0 = physical.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'direction': 'input'})path0.width
        >>> pin0.width 
        100

        Notes
        -----
        Reference in Korean:
        int: 객체의 폭.
        """
        return abs(self.xy[0, 0] - self.xy[1, 0])

    @property
    def size(self):
        """attribute
        numpy.ndarray: Object size.

        

        Examples
        --------
        >>> pin0 = physical.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'direction': 'input'})path0.width
        >>> pin0.size 
        [100, 100]

        Notes
        -----
        Reference in Korean:
        numpy.ndarray: 객체의 크기.
        """
        return np.array([self.width, self.height])

    @property
    def height_vec(self):
        """numpy.ndarray(dtype=int): Return np.array([0, height])."""
        return np.array([0, self.height])

    @property
    def width_vec(self):
        """numpy.ndarray(dtype=int): Return np.array([width, 0])."""
        return np.array([self.width, 0])

    def __init__(self, xy, layer, name=None, netname=None, params=None, master=None, elements=None):
        """
        Constructor function of Pin class.

        Parameters
        ----------
        xy : numpy.ndarray
            physical coordinates of object, bbox.
        layer : list
            layer information of object.
        name : str
            object name.
        netname : str
            node name of object.
        params : dict
            dictionary having attributes.
        
        Returns
        -------
        laygo2.Pin

        See Also
        --------
        Class IterablePhysicalObject

        Examples
        --------
        >>> pin0 = physical.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'direction': 'input'})path0.width
        >>> print( pin0 ) 
        <laygo2.object.physical.Pin object at 0x0000017DBDA4F508> name: None,
         class: Pin, xy: [[0, 0], [100, 100]], params: {'direction': 'input’},
         layer: ['M1', 'drawing'], netname: net0, shape: None, master: None

        Notes
        -----
        Reference in Korean:
        Pin 클래스의 생성자 함수
        파라미터
        xy(numpy.ndarray): 객체의 물리적 좌표, bbox
        layer(list): 객체의 layer 정보
        name(str): 객체의 이름
        netname(str): 객체의 노드 명
        params(dict): 속성을 갖는 Dictionary
        반환값
        laygo2.Pin
        참조
        Class IterablePhysicalObject
        """
        self.layer = np.asarray(layer)
        if netname is None:
            netname = name
        self.netname = netname
        self.master = master
        IterablePhysicalObject.__init__(self, xy=xy, name=name, params=params, elements=elements)

    def summarize(self):
        """Return the summary of the object information."""
        return IterablePhysicalObject.summarize(self) + ", " + \
                                                      "layer: " + str(self.layer) + ", " + \
                                                      "netname: " + str(self.netname) + ", " + \
                                                      "shape: " + str(self.shape) + ", " + \
                                                      "master: " + str(self.master)

    def export_to_dict(self):
        db = dict()
        db['xy'] = self.xy.tolist()
        db['layer'] = self.layer.tolist()
        db['name'] = self.name
        db['netname'] = self.netname
        return db

class Text(PhysicalObject):
    """
    Text object class.

    Attributes
    ----------
    layer : list
    text : str

    Methods
    -------
    __init__()

    Notes
    -----
    Reference in Korean:
    Text 객체 클래스.
    """
    layer = None
    """attribute
    numpy.ndarray: Layer information of object.

    

    Examples
    --------
    >>> text0 = physical.Text(xy=[[ 0, 0], [100,100 ]], layer=['text', 'drawing'], text='test', params=None)
    >>> text0.layer 
    ['text', 'drawing']

    Notes
    -----
    Reference in Korean:
    numpy.ndarray: 객체의 Layer 정보.
    """
    text = None
    """attribute
    str: Text content of object.

    

    Examples
    --------
    >>> text0 = physical.Text(xy=[[ 0, 0], [100,100 ]], layer=['text', 'drawing'], text='test', params=None)
    >>> text0.text 
    ”text”

    Notes
    -----
    Reference in Korean:
    str: 객체의 텍스트 내용.
    """

    def __init__(self, xy, layer, text, name=None, params=None):
        """
        Constructor function of Text class.

        Parameters
        ----------
        xy : numpy.ndarray
            physical coordinates of object, bbox.
        layer : list
            layer information of object.
        text : str
            text content.
        name : str
            object name.
        params : dict
            dictionary having attributes.
        
        Returns
        -------
        laygo2.Text

        See Also
        --------
        Class PhysicalObject

        Examples
        --------
        >>> pin0 = physical.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'direction': 'input'})path0.width
        >>> pin0.text 
        [100, 100]

        Notes
        -----
        Reference in Korean:
        Text 클래스의 생성자 함수
        파라미터
        xy(numpy.ndarray): 객체의 물리적 좌표, bbox
        layer(list): 객체의 layer 정보
        text(str): 텍스트 내용
        name(str): 객체의 이름
        params(dict): 속성을 갖는 Dictionary
        반환값
        laygo2.Text
        참조
        Class PhysicalObject
        """
        self.layer = layer
        self.text = text

        PhysicalObject.__init__(self, xy=xy, name=name, params=params)

    def summarize(self):
        """Return the summary of the object information."""
        return PhysicalObject.summarize(self) + ", " + \
                                              "layer: " + str(self.layer) + ", " + \
                                              "text: " + str(self.text)

class Instance(IterablePhysicalObject):
    """
    Instance object class.

    Attributes
    ----------
    libname : str
    cellname : str
    unit_size : numpy.ndarray
    transform : str
    pins : dict
    xy0 : numpy.ndarray
    xy1 : numpy.ndarray
    size : numpy.ndarray
    pitch : numpy.ndarray
    spacing : numpy.ndarray
    height : int
    width : int

    Methods
    -------
    __init__()
    ndenumerate()

    Notes
    -----
    Reference in Korean:
    Instance 객체 클래스.
    """
    # TODO: update (maybe) xy and sub-elements after transform property is updated.

    libname = None
    """attribute
    str: Library name of object.

    

    Examples
    --------
    >>> inst0_pins = dict() 
    >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
    >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform = 'R0’)
    >>> inst0.libname 
    mylib

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_libname.png

    Reference in Korean:
    str: 객체의 라이브러리 이름.
    """

    cellname = None
    """attribute
    str: Name of the cell to which the object belongs.

    

    Examples
    --------
    >>> inst0_pins = dict() 
    >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
    >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
    >>> inst0.cellname
    mycell

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_cellname.png

    Reference in Korean:
    str: 객체가 속한 셀 이름.
    """

    viewname = None
    """str: The view name of the instance."""

    shape = None
    """np.array([int, int]) or None: The shape of the instance mosaic. None if the instance is non-mosaic."""

    _pitch = None
    """np.array([int, int]) or None: The internal variable for pitch."""

    unit_size = None
    """attribute
    numpy.ndarray: Unit size when the object is in array.

    

    Examples
    --------
    >>> inst0_pins = dict() 
    >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
    >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
    >>> inst0.unit_size 
    [100, 100]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_unit_size.png

    Reference in Korean:
    numpy.ndarray: 객체가 배열로 구성되었을 때 단위 크기.
    """
    transform = 'R0'
    """attribute
    str: Transformation attribute of object.

    

    Examples
    --------
    >>> inst0_pins = dict() 
    >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
    >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
    >>> inst0.transform 
    “R0”

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_transform.png

    Reference in Korean:
    str: 객체의 변환 속성.
    """

    pins = None
    """attribute
    dict: Dictionary having the pins belonging to object.

    

    Examples
    --------
    >>> inst0_pins = dict() 
    >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
    >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
    >>> inst0.pins 
    {'in': <laygo2.object.physical.Pin object at 0x000001CA76EE1348>, 'out': <laygo2.object.physical.Pin object at 0x000001CA7709BD48>} 
    >>> inst0.pins[“in”].shape 
    [3,2]
    >>> inst0.pins[“out”].shape 
    [3,2] 
    >>> inst0.pins[“in”][1,1].xy 
    [ [300, 300], [310, 310] ]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_pins.png

    Reference in Korean:
    dict: 객체에 속한 핀들을 갖는 Dict.
    """

    def _update_pins(self, xy_ofst):
        """Update xy-coordinates of this object's pins. An internal function for _set_xy()"""
        if self.pins is not None:
            for pn, p in self.pins.items():
                if np.all(p is not None):
                    # Update the x and y coordinate values of elements.
                    for n, e in np.ndenumerate(p):
                        if e is not None:
                            e.xy = e.xy + xy_ofst

    def _get_xy(self):
        """numpy.ndarray(dtype=numpy.int): Get the x and y coordinate values of this object."""
        return self._xy

    def _set_xy(self, value):
        """numpy.ndarray(dtype=numpy.int): Set the x and y coordinate values of this object."""
        # Update the coordinate value of its pins.
        self._update_pins(xy_ofst=value - self.xy)
        IterablePhysicalObject._set_xy(self, value=value)

    xy = property(_get_xy, _set_xy)

    @property
    def xy0(self):
        """attribute
        numpy.ndarray: Coordinates of major corner of object.

        

        Examples
        --------
        >>> inst0_pins = dict() 
        >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
        >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
        >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
        >>> inst0.xy0 
        [100, 100]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_xy0.png

        Reference in Korean:
        numpy.ndarray: 객체의 주 코너 좌표.
        """
        return self.xy

    @property
    def xy1(self):
        """attribute
        numpy.ndarray: Coordinates of minor corner of object.

        

        Examples
        --------
        >>> inst0_pins = dict() 
        >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
        >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
        >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
        >>> inst0.xy1 
        [600, 400]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_xy1.png

        Reference in Korean:
        numpy.ndarray: 객체의 보조 코너 좌표.
        """
        if self.size is None:
            return self.xy
        else:
            return self.xy + np.dot(self.size, tf.Mt(self.transform).T)

    @property
    def size(self):
        """attribute
        numpy.ndarray: Object size.

        

        Examples
        --------
        >>> inst0_pins = dict() 
        >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
        >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
        >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
        >>> inst0.size 
        [500, 300]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_size.png

        Reference in Korean:
        numpy.ndarray: 객체의 크기.
        """
        if self.shape is None:
            return self.unit_size
        else:
            return (self.shape - np.array([1, 1])) * self.pitch + self.unit_size

    def get_pitch(self):
        """numpy.ndarray(dtype=int): Get the pitch of the instance."""
        if self._pitch is None:
            return self.unit_size
        else:
            return self._pitch

    def set_pitch(self, value):
        """numpy.ndarray(dtype=int): Set the pitch of the instance."""
        self._pitch = value

    pitch = property(get_pitch, set_pitch)
    """attribute
    numpy.ndarray: Pitch between unit object of the object in array.

    

    Examples
    --------
    >>> inst0_pins = dict() 
    >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
    >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
    >>> inst0.pitch 
    [200, 200]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_pitch.png

    Reference in Korean:
    numpy.ndarray: 배열로 구성된 객체의 단위객체간 간격.
    """

    def get_spacing(self):
        return self.pitch

    def set_spacing(self, value):
        self.pitch = value

    spacing = property(get_spacing, set_spacing)
    """attribute
    numpy.ndarray: Spacing between unit object of the object in array.

    

    Examples
    --------
    >>> inst0_pins = dict() 
    >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
    >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
    >>> inst0.spacing 
    [200, 200]

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_spacing.png

    Reference in Korean:
    numpy.ndarray: 배열로 구성된 객체의 단위객체간 간격.
    """

    @property
    def bbox(self):
        bbox = np.array([self.xy0, self.xy1])
        #return bbox
        #return self.xy + np.dot(self.size, tf.Mt(self.transform).T)
        return np.sort(bbox, axis=0)

    @property
    def height(self):
        """attribute
        int: Object height

        

        Examples
        --------
        >>> inst0_pins = dict() 
        >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
        >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
        >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
        >>> inst0.height 
        300

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_height.png

        Reference in Korean:
        int: 객체의 높이.
        """
        return abs(self.bbox[1][1] - self.bbox[0][1])

    @property
    def width(self):
        """attribute
        int: Object width

        

        Examples
        --------
        >>> inst0_pins = dict() 
        >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
        >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
        >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
        >>> inst0.width 
        500

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_width.png

        Reference in Korean:
        int: 객체의 폭.
        """
        return abs(self.bbox[1][0] - self.bbox[0][0])

    @property
    def height_vec(self):
        """numpy.ndarray(dtype=int): Return np.array([0, height])."""
        return np.array([0, self.height])

    @property
    def width_vec(self):
        """numpy.ndarray(dtype=int): Return np.array([width, 0])."""
        return np.array([self.width, 0])

    def __init__(self, xy, libname, cellname, viewname='layout', shape=None, pitch=None, transform='R0',
                 unit_size=np.array([0, 0]), pins=None, name=None, params=None):
        """
        Constructor function of Instance class.

        Parameters
        ----------
        xy : numpy.ndarray
            major coordinates of object. [x0, y0]
        libname : str
            library name of object.
        cellname : str
            cell name of object.
        shape : numpy.ndarray
            shape of elements. [column, row]
        pitch : numpy.ndarray
            pitch between lower members of the object in array.
        transform : str
            transformation attribute of the object.
        unit_size : list
            unit size of object.
        pins : dict
            dictionary having pins belonging to the object.
        name : str
            object name.
        params : dict
            dictionary having attributes of the object.
        
        Returns
        -------
        object (laygo2.Instance)

        See Also
        --------
        Class IterablePhysicalObject

        Examples
        --------
        >>> inst0_pins = dict() 
        >>> inst0_pins[‘in’] = physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in’) 
        >>> inst0_pins[‘out’]= physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
        >>> inst0 = physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell", shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins , transform =  'R0’)
        >>> print( inst0[1,0].xy0 ) 
        [ 300, 100 ]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_instance_init.png

        Reference in Korean:
        Instance 클래스의 생성함수
        파라미터
        xy(numpy.ndarray): 객체의 주좌표 [x0, y0]
        libname(str): 객체의 library 이름
        cellname(str): 객체의 cell이름
        shape(numpy.ndarray): elements 의 shape, [column, row]
        pitch(numpy.ndarray): 배열로 구성된 객체의 하위객체간 간격
        transform(str): 객체의 변환 속성
        unit_size(list): 객체의 단위 크기
        pins(dict): 객체에 속한 핀들을 갖는 Dictionary
        name(str): 객체의 이름
        params(dict): 객체의 속성을 갖는 Dictionary
        반환값
        object (laygo2.Instance)
        참조
        Class IterablePhysicalObject
        """
        # Assign parameters.
        xy = np.asarray(xy)
        self.libname = libname
        self.cellname = cellname
        self.viewname = viewname
        if shape is not None:
            _shape = np.asarray(shape)
            if _shape.shape != (2, ):
                raise ValueError('Instance shape should be a (2, ) numpy array or None.')
            self.shape = _shape
        if pitch is not None:
            self.pitch = np.asarray(pitch)
        self.transform = transform
        self.unit_size = np.asarray(unit_size)

        # Construct an array for elements.
        if shape is None:
            # elements = self  # self-referencing causes recursion errors. Deprecated.
            elements = None
        else:
            _shape = tuple(shape)
            elements = np.zeros(_shape, dtype=np.object)
            # elements = LayoutObjectArray(np.zeros(_shape, dtype=np.object))
            _it = np.nditer(elements, flags=['multi_index', 'refs_ok'])
            while not _it.finished:
                _idx = _it.multi_index
                _xy = xy + np.dot(self.pitch * np.array(_idx), tf.Mt(self.transform).T)
                inst = Instance(xy=_xy, libname=libname, cellname=cellname, shape=None, pitch=pitch,
                                transform=self.transform, unit_size=self.unit_size, pins=pins, name=name, params=params)
                elements[_idx] = inst
                _it.iternext()

        IterablePhysicalObject.__init__(self, xy=xy, name=name, params=params, elements=elements)
        # Create the pin dictionary. Can we do the same thing without generating these many Pin objects?
        self.pins = dict()
        if pins is not None:
            if not isinstance(pins, dict):
                raise ValueError("The pins parameter for Instance objects should be a dictionary.")
            for pn, p in pins.items():
                _xy0 = xy + np.dot(p.xy, tf.Mt(transform).T)
                if shape is not None:
                    elements = []
                    for i in range(shape[0]):
                        elements.append([])
                        for j in range(shape[1]):
                            _xy = _xy0 + np.dot(self.pitch * np.array([i, j]), tf.Mt(transform).T)
                            # If p has elements, they need to be copied and transferred to the new pin.
                            _pelem = None
                            if p.elements is not None:
                                _pelem = np.empty(p.elements.shape, dtype=object)
                                for _idx, _pe in np.ndenumerate(p.elements):
                                    _pexy0 = xy + np.dot(_pe.xy, tf.Mt(transform).T) \
                                             + np.dot(self.pitch * np.array([i, j]), tf.Mt(transform).T)
                                    _pelem[_idx] = Pin(xy=_pexy0, netname=_pe.netname, layer=_pe.layer, name=_pe.name, master=self)
                            pin = Pin(xy=_xy, netname=p.netname, layer=p.layer, name=p.name, master=self,
                                      elements=_pelem)  # master uses self instead of self.elements[i, j].
                            elements[i].append(pin)
                    elements = np.array(elements)
                else:
                    # If p has elements, they need to be copied and transferred to the new pin.
                    _pelem = None
                    if p.elements is not None:
                        _pelem = np.empty(p.elements.shape, dtype=object)
                        for _idx, _pe in np.ndenumerate(p.elements):
                            _pexy0 = xy + np.dot(_pe.xy, tf.Mt(transform).T)
                            _pelem[_idx] = Pin(xy=_pexy0, netname=_pe.netname, layer=_pe.layer, name=_pe.name, master=self)
                    elements = _pelem
                self.pins[pn] = Pin(xy=_xy0, netname=p.netname, layer=p.layer, name=p.name, master=self,
                                    elements=elements)

    def summarize(self):
        """Summarize object information."""
        return PhysicalObject.summarize(self) + ", " + \
               "size: " + str(self.size.tolist()) + ", " + \
               "shape: " + str(None if self.shape is None else self.shape.tolist()) + ", " + \
               "pitch: " + str(self.pitch.tolist()) + ", " + \
               "transform: " + str(self.transform) + ", " + \
               "pins: " + str(self.pins)

class VirtualInstance(Instance):  # IterablePhysicalObject):
    """
    VirtualInstance object class.

    Attributes
    ----------
    native_elemnets : dict

    Methods
    -------
    __init__()
    ndenumerate()

    Notes
    -----
    Reference in Korean:
    VirtualInstance 객체 클래스.
    """

    native_elements = None
    """attribute
    dict: Dictionary having physical entities constituting the object.
        Dictionary consisting of physicalObject.
    
    

    Examples
    --------
    >>> vinst0_pins = dict() 
    >>> vinst0_pins['in']  = physical.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in') 
    >>> vinst0_pins['out'] = physical.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out’)
    >>> vinst0_native_elements = dict() 
    >>> vinst0_native_elements['R0'] = physical.Rect(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing']) 
    >>> vinst0_native_elements['R1'] = physical.Rect(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing']) 
    >>> vinst0_native_elements['R2'] = physical.Rect(xy=[[0, 0], [100, 100]], layer=['prBoundary', 'drawing’])
    >>> vinst0 = physical.VirtualInstance(name='I0', libname='mylib', cellname='myvcell', xy=[500, 500], native_elements=vinst0_native_elements, shape=[3, 2], pitch=[100, 100], unit_size=[100, 100], pins=vinst0_pins, transform='R0’)
    >>> vinst0.native_elements 
    { ‘R0’ : physical.Rect, ‘R1’ : physical.Rect, ‘R2’ : physical.Rect }

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_VirtualInstance_native_elements.png

    Reference in Korean:
    dict: 객체를 구성하는 물리객체들을 갖고있는 Dict.
        PhysicalObject로 구성된 Dictionary.
    """
    # Dict[PhysicalObject] the elements that compose the virtual instance. Its keys represent the names of the elements.

    def __init__(self, xy, libname, cellname, native_elements, viewname='layout', shape=None, pitch=None,
                 transform='R0', unit_size=np.array([0, 0]), pins=None, name=None, params=None):
        """
        Constructor function of VirtualInstance class.

        Parameters
        ----------
        xy : numpy.ndarray
            major coordinates of object [x0, y0].
        libname : str
            library name of object.
        cellname : str
            cell name of object.
        native_elements : dict
            dictionary having physical entities constituting the object.
        shape : numpy.ndarray
            shape of elements.
        pitch : numpy.ndarray
            pitch between lower members of the object in array.
        transform : str
            transformation attribute of the object.
        unit_size : list
            unit size of object.
        pins : dict
            dictionary having pins belonging to the object.
        name : str
            object name.
        params : dict
            dictionary having attributes of the object.

        Returns
        -------
        object (laygo2.VirtualInstance)

        See Also
        --------
        Class Instance

        Examples
        --------
        >>> vinst0 = physical.VirtualInstance(name='I0', libname='mylib', cellname='myvcell', xy=[500, 500], native_elements=vinst0_native_elements, shape=[3, 2], pitch=[100, 100], unit_size=[100, 100], pins=vinst0_pins, transform='R0’)
        >>> vinst0.native_elements 
        { ‘R0’ : physical.Rect, ‘R1’ : physical.Rect, ‘R2’ : physical.Rect }

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_physical_VirtualInstance_init.png

        Reference in Korean:
        VirtualInstance 클래스의 생성자 함수
        파라미터
        xy(numpy.ndarray): 객체의 주좌표 [x0, y0]
        libname(str): 객체의 library 이름
        cellname(str): 객체의 cell이름
        native_elements(dict): 객체를 구성하는 물리객체를 갖는 Dict
        shape(numpy.ndarray): elements 의 shape
        pitch(numpy.ndarray): 배열로 구성된 객체의 하위객체간 간격 
        transform(str): 객체의 변환 속성
        unit_size(list): 객체의 단위 크기
        pins(dict): 객체에 속한 핀들을 갖는 Dictionary
        name(str): 객체의 이름
        params(dict): 객체의 속성을 갖는 Dictionary
        반환값
        object (laygo2.VirtualInstance)
        참조
        Class Instance
        """
        self.native_elements = native_elements

        Instance.__init__(self, xy=xy, libname=libname, cellname=cellname, viewname=viewname, shape=shape, pitch=pitch,
                          transform=transform, unit_size=unit_size, pins=pins, name=name, params=params)
        #Instance.__init__(self, xy=xy, libname='VirtualInstance', cellname='VirtualInstance', shape=shape, pitch=pitch,
        #                  transform=transform, unit_size=unit_size, pins=pins, name=name, params=params)

    def summarize(self):
        """Summarize object information."""
        return Instance.summarize(self) + ", " + \
               "native elements: " + str(self.native_elements)

    def get_element_position(self, obj ):
        """
        Get element's xy-position from origin
        
        Parameters
        ----------
        obj : element
            element belongs to self
        """
        vinst = self
        tr    = vinst.transform
        coners = np.zeros((4, 2))
        v_r = np.zeros(2)  # for rotation
        bbox_raw = obj.bbox
        offset = vinst.xy
        if tr == "R0":
            v_r = v_r + (1, 1)
            coners[0] = offset + v_r * bbox_raw[0]
            coners[2] = offset + v_r * bbox_raw[1]
        elif tr == "MX":
            v_r = v_r + (1, -1)
            coners[1] = offset + v_r * bbox_raw[0]
            coners[3] = offset + v_r * bbox_raw[1]
            coners[0] = coners[0] + (coners[1][0], coners[3][1])
            coners[2] = coners[2] + (coners[3][0], coners[1][1])
        elif tr == "MY":
            v_r = v_r + (-1, 1)
            coners[3] = offset + v_r * bbox_raw[0]
            coners[1] = offset + v_r * bbox_raw[1]
            coners[0] = coners[0] + (coners[1][0], coners[3][1])
            coners[2] = coners[2] + (coners[3][0], coners[1][1])
        elif tr == "R90":
            v_r = v_r + (-1, -1)
            coners[2] = offset + v_r * bbox_raw[0]
            coners[0] = offset + v_r * bbox_raw[1]
        else:
            raise ValueError(" Others transfom not implemented")
        return coners[0], coners[2]








# Test
if __name__ == '__main__':
    test_rect = False
    test_path = False
    test_pin = False
    test_text = False
    test_pointer = False
    test_instance = True
    test_virtual_instance = False

    # You can create various objects by running part of the following commands.
    if test_rect:
        print("Rect test")
        rect0 = Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'maxI': 0.005})
        print(rect0)
    if test_path:
        print("Path test")
        path0 = Path(xy=[[0, 0], [0, 100]], width=10, extension=5, layer=['M1', 'drawing'], netname='net0')
        print(path0)
    if test_pin:
        print("Pin test")
        pin0 = Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', master=rect0,
                   params={'direction': 'input'})
        print(pin0)
    if test_text:
        print("Text test")
        text0 = Text(xy=[0, 0], layer=['text', 'drawing'], text='test', params=None)
        print(text0)
    if test_instance:
        print("Instance test - creating a vanilla instance.")
        inst0_pins = dict()
        inst0_pins['in'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
        inst0_pins['out'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
        inst0 = Instance(name='I0', xy=[100, 100], libname='mylib', cellname='mycell', shape=[3, 2], pitch=[100, 100],
                         unit_size=[100, 100], pins=inst0_pins, transform='R0')
        print("  ", inst0)
        print("  ", inst0.pointers)
        print(inst0.elements)
        for idx, it in inst0.ndenumerate():
            print("what?")
            print("  ", idx, it)
            print("  ", idx, it.pins['in'])
        print("Instance test - updating the instance's coordinate values.")
        inst0.xy = [200, 200]
        print("  ", inst0)
        print("  ", inst0.pointers)
        for idx, it in inst0.ndenumerate():
            print("  ", idx, it)
            print("  ", idx, it.pins['in'])
    if test_virtual_instance:
        print("VirtualInstance test - creating a vanilla virtual instance.")
        inst1_pins = dict()
        inst1_pins['in'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
        inst1_pins['out'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
        inst1_native_elements = dict()
        inst1_native_elements['R0'] = Rect(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'])
        inst1_native_elements['R1'] = Rect(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'])
        inst1_native_elements['R2'] = Rect(xy=[[0, 0], [100, 100]], layer=['prBoundary', 'drawing'])
        inst1 = VirtualInstance(name='I0', libname='mylib', cellname='myvcell', xy=[500, 500],
                                native_elements=inst1_native_elements, shape=[3, 2], pitch=[100, 100],
                                unit_size=[100, 100], pins=inst1_pins, transform='R0')
        print("  ", inst1)
        for idx, it in inst1.ndenumerate():
            print("  ", idx, it.pins['in'])
        for idx, it in inst1.pins['in'].ndenumerate():
            print("  ", idx, it)



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements classes for various layout objects.
"""

__author__ = "Jaeduk Han"
__maintainer__ = "Jaeduk Han"
__status__ = "Prototype"

import numpy as np
import laygo2.util.transform as tf


class Pointer: #LayoutObject):
    """
    Pointer class. Pointers are attached to other LayoutObjects to indicate various positional information
    associated with the object for placing and routing purposes.

    See Also
    --------
    LayoutObject
    """

    name = None
    """str or None: the name of the object."""

    _xy = np.zeros(2, dtype=np.int)
    """numpy.ndarray(dtype=numpy.int): the internal variable of xy."""

    def get_xy(self):
        """numpy.ndarray(dtype=int): gets the xy."""
        return self._xy

    def set_xy(self, value):
        """numpy.ndarray(dtype=int): sets the xy."""
        self._xy = np.asarray(value, dtype=np.int)

    xy = property(get_xy, set_xy)
    """numpy.ndarray(detype=numpy.int): the xy-coordinates of the object."""

    direction = None
    """str: the direction of the pointer."""
    master = None
    """LayoutObject.Instance: the master instance that the point tag is attached."""
    params = None
    """dict or None: the parameters of the object. """


    def __init__(self, name, xy, direction=None, master=None, params=None):
        """
        Constructor.

        Parameters
        ----------
        name : str
            The name of the object.
        xy : numpy.ndarray(dtype=int)
            The xy-coordinates of the object. Its default format is [x, y].
        direction : str
            The direction of the pointer, which will be used for place functions.
        master : LayoutObject.LayoutObject or its derived classes
            The master instance handle.
        """
        self.name = name
        self.xy = xy
        self.params = params
        self.direction = direction
        self.master = master

    def __str__(self):
        return self.summarize()

    def summarize(self):
        """Returns the summary of the object information."""
        return self.__repr__() + " " \
                                 "name: " + self.name + ", " + \
                                 "class: " + self.__class__.__name__ + ", " + \
                                 "xy: " + str(self.xy.tolist()) + ", " + \
                                 "params: " + str(self.params) + ", " + \
                                 "direction: " + str(self.direction) + ", " + \
                                 "master: /% " + str(self.master) + " %/"


class LayoutObject:
    """
    Basic layout object.
    """

    name = None
    """str or None: the name of the object."""

    _xy = np.zeros(2, dtype=np.int)
    """numpy.ndarray(dtype=numpy.int): the internal variable of xy."""

    def get_xy(self):
        """numpy.ndarray(dtype=int): gets the xy."""
        return self._xy

    def set_xy(self, value):
        """numpy.ndarray(dtype=int): sets the xy."""
        self._xy = np.asarray(value, dtype=np.int)
        self.update_pointers()

    xy = property(get_xy, set_xy)
    """numpy.ndarray(detype=numpy.int): the xy-coordinates of the object."""

    @property
    def bbox(self):
        """numpy.ndarray(dtype=int): the bounding box of the object. Its default format is
        [[x_ll, y_ll], [x_ur, y_ur]]"""
        return np.sort(np.array([self.xy, self.xy]), axis=0)

    params = None
    """dict or None: the parameters of the object. """

    # Pointers
    pointers = dict()

    def __init__(self, name, xy, params=None):
        """
        Constructor.

        Parameters
        ----------
        name : str
            The name of the object.
        xy : numpy.ndarray([int, int])
            The xy-coordinates of the object.
        params : dict or None
            Additional parameters of the object.
        """
        self.name = name

        # initialize pointers
        self.pointers['left'] = Pointer(name='left', xy=[0, 0], direction='left', master=self)
        self.pointers['right'] = Pointer(name='right', xy=[0, 0], direction='right', master=self)
        self.pointers['bottom'] = Pointer(name='bottom', xy=[0, 0], direction='bottom', master=self)
        self.pointers['top'] = Pointer(name='top', xy=[0, 0], direction='top', master=self)
        self.pointers['bottom_left'] = Pointer(name='bottom_left', xy=[0, 0], direction='bottom_left', master=self)
        self.pointers['bottom_right'] = Pointer(name='bottom_right', xy=[0, 0], direction='bottom_right', master=self)
        self.pointers['top_left'] = Pointer(name='top_left', xy=[0, 0], direction='top_left', master=self)
        self.pointers['top_right'] = Pointer(name='top_right', xy=[0, 0], direction='top_right', master=self)
        self.left = self.pointers['left']
        self.right = self.pointers['right']
        self.bottom = self.pointers['bottom']
        self.top = self.pointers['top']
        self.bottom_left = self.pointers['bottom_left']
        self.bottom_right = self.pointers['bottom_right']
        self.top_left = self.pointers['top_left']
        self.top_right = self.pointers['top_right']

        self.params = params
        self.xy = xy

    def __str__(self):
        return self.summarize()

    def summarize(self):
        """Returns the summary of the object information."""
        return self.__repr__() + " " \
                                 "name: " + self.name + ", " + \
                                 "class: " + self.__class__.__name__ + ", " + \
                                 "xy: " + str(self.xy.tolist()) + ", " + \
                                 "params: " + str(self.params)

    def update_pointers(self):
        xy_left = np.diag(np.dot(np.array([[1, 0], [0.5, 0.5]]), self.bbox)).astype(np.int)
        xy_right = np.diag(np.dot(np.array([[0, 1], [0.5, 0.5]]), self.bbox)).astype(np.int)
        xy_bottom = np.diag(np.dot(np.array([[0.5, 0.5], [1, 0]]), self.bbox)).astype(np.int)
        xy_top = np.diag(np.dot(np.array([[0.5, 0.5], [0, 1]]), self.bbox)).astype(np.int)
        xy_bottom_left = np.diag(np.dot(np.array([[1, 0], [1, 0]]), self.bbox)).astype(np.int)
        xy_bottom_right = np.diag(np.dot(np.array([[0, 1], [1, 0]]), self.bbox)).astype(np.int)
        xy_top_left = np.diag(np.dot(np.array([[1, 0], [0, 1]]), self.bbox)).astype(np.int)
        xy_top_right = np.diag(np.dot(np.array([[0, 1], [0, 1]]), self.bbox)).astype(np.int)
        self.pointers['left'].xy = xy_left
        self.pointers['right'].xy = xy_right
        self.pointers['bottom'].xy = xy_bottom
        self.pointers['top'].xy = xy_top
        self.pointers['bottom_left'].xy = xy_bottom_left
        self.pointers['bottom_right'].xy = xy_bottom_right
        self.pointers['top_left'].xy = xy_top_left
        self.pointers['top_right'].xy = xy_top_right



class LayoutIterableObject(LayoutObject):
    """
    Layout object that contains iterable elements. The iteration feature is implemented through a numpy ndarray object,
    called elements, by mapping iterator-related functions. Indexing sub-elements follows the numpy convention, which
    provides easy multi-dimensional indexing and advanced slicing.

    See Also
    --------
    LayoutObject
    """

    elements = None
    """numpy.array(dtype=LayoutObject): the iterable elements."""

    def __getitem__(self, pos):
        """Returns its sub-elements based on pos parameter."""
        return self.elements[pos]

    def __setitem__(self, key, item):
        self.elements[key] = item

    def __iter__(self):
        """Iterator function. Directly mapped to its elements."""
        return self.elements.__iter__()

    def __next__(self):
        """Iterator function. Directly mapped to its elements."""
        return self.elements.__next__()

    def ndenumerate(self):
        """Enumerates over the element array. Calls np.ndenumerate() of its elements."""
        return np.ndenumerate(self.elements)

    def __init__(self, name, xy=None, params=None, elements=None):
        """
        Constructor.

        Parameters
        ----------
        name : str
            The name of the object.
        xy : numpy.ndarray(dtype=int)
            The xy-coordinates of the layout object. Its default format is [x, y].
        params : dict or None
            The parameters of the object.
        elements : numpy.ndarray(dtype=LayoutObject) or None
            The iterable elements of the object.
        """

        self.elements = elements
        LayoutObject.__init__(self, name=name, xy=xy, params=params)


class LayoutObjectArray(np.ndarray):
    """LayoutObject array class for containing multiple layout objects. Subclassing ndarray to utilize advance slicing
     functions."""
    name = None
    """str: the name of the object."""

    params = None
    """dict or None: parameters of the object. """

    def __new__(cls, input_array, name=None, xy=None, params=None):
        """
        Constructor for ndarray subclasses - check the NumPy manual for details.

        Parameters
        ----------
        input_array : np.ndarray
            An array of LayoutObject objects.
        name : str
            The name of the array.
        xy : numpy.ndarray(dtype=int)
            The xy-coordinate of the object. The format is [x0, y0].
        params : dict
            Additional parameters of the array.
        """
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj.name = name
        obj.xy = xy
        obj.params = params
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        """
        Array finalizing function for subclassing ndarray - check the NumPy manual for details
        """
        if obj is None: return
        # Transfer parameters
        self.name = getattr(obj, 'name', None)
        self.xy = getattr(obj, 'xy', None)
        self.params = getattr(obj, 'params', None)

    def summarize(self):
        """Summarizes object information."""
        return "  " + \
               "name:" + self.name + ", " + \
               "class:" + self.__class__.__name__ + ", " + \
               "shape:" + str(self.shape) + ", " + \
               "xy:" + str(self.xy) + ", " + \
               "params:" + str(self.params) + "\n"


class Rect(LayoutObject):
    """
    Rect object.

    See Also
    --------
    LayoutObject
    """

    layer = None
    """list(str): Rect layer. The format is [name, purpose]."""
    netname = None
    """str: the name of the net associated with the rect."""

    def get_xy0(self):
        """Gets the xy0."""
        return self.xy[0]

    def set_xy0(self, value):
        """Sets the xy0."""
        self.xy[0] = np.asarray(value)
        self.update_pointers()

    xy0 = property(get_xy0, set_xy0)
    """numpy.array(dtype=int): the xy-coordinate of the primary (mostly lower-left) corner of the rect."""

    def get_xy1(self):
        """gets the xy1."""
        return self.xy[1]

    def set_xy1(self, value):
        """Sets the xy1."""
        self.xy[1] = np.asarray(value)
        self.update_pointers()

    xy1 = property(get_xy1, set_xy1)
    """numpy.array(dtype=int): the xy-coordinate of the secondary (mostly upper-right) corner of the rect."""

    @property
    def height(self):
        """int: the height of the rect"""
        return abs(self.xy0[1] - self.xy1[1])

    @property
    def width(self):
        """int: the width of the rect"""
        return abs(self.xy0[0] - self.xy1[0])

    @property
    def size(self):
        """numpy.ndarray(dtype=int): the size of the rect."""
        return np.array([self.width, self.height])

    @property
    def bbox(self):
        """numpy.ndarray(dtype=int): the bounding box of the object. Its default format is
        [[x_ll, y_ll], [x_ur, y_ur]]"""
        return np.sort(np.array([self.xy0, self.xy1]), axis=0)

    pointers = dict()
    """dict(): pointer dictionary."""
    # frequently used pointers
    left = None
    right = None
    top = None
    bottom = None
    center = None
    bottom_left = None
    bottom_right = None
    top_left = None
    top_right = None

    def __init__(self, name, xy, layer, netname=None, params=None):
        """
        Constructor.

        Parameters
        ----------
        name : str
            The name of the object
        xy : numpy.ndarray(dtype=int)
            The xy-coordinates of the object. Its format is [[x0, y0], [x1, y1]].
        layer : list(str)
            The layer information of the object. Its format is [layer, purpose]
        netname : str, optional
            The name of net associated with the object.
        params : dict
            Additional parameters of the object.
        """
        LayoutObject.__init__(self, name, xy, params=params)

        self.layer = layer
        self.netname = netname

    def summarize(self):
        """Returns the summary of the object information."""
        return LayoutObject.summarize(self) + ", " + \
                                              "layer: " + str(self.layer) + ", " + \
                                              "netname: " + str(self.netname)


class Wire(Rect):
    """
    Wire object.

    See Also
    --------
    LayoutObject
    """
    width = None
    """int: the width of the wire"""
    extension = 0
    """int: the amount of extension from the both edges of the wire."""

    def __init__(self, name, xy, layer, width, extension=0, netname=None, params=None):
        self.width = width
        self.extension = extension

        Rect.__init__(self, name, xy, layer, netname=netname, params=params)

    def summarize(self):
        """Returns the summary of the object information."""
        return LayoutObject.summarize(self) + ", " + \
               "width: " + str(self.width) + ", " + \
               "extension: " + str(self.extension) + ", " + \
               "layer: " + str(self.layer) + ", " + \
               "netname: " + str(self.netname)


class Pin(LayoutIterableObject):
    """
    Pin object.

    See Also
    --------
    LayoutIterableObject
    """

    layer = None
    """list(str): the layer of the pin. Its format is [name, purpose]."""
    netname = None
    """str: the name of the net associated with the pin."""
    master = None
    """Instance: the master instance of the pin. Used for instance pins only."""
    elements = None
    """numpy.array(dtype=Pin): the array contains its sub-element pins."""

    def get_xy0(self):
        """numpy.ndarray(dtype=int): gets the primary corner of the instance."""
        return self.xy[0]

    def set_xy0(self, value):
        """numpy.ndarray(dtype=int): sets the primary corner of the instance."""
        self.xy[0] = np.asarray(value)
        self.update_pointers()

    xy0 = property(get_xy0, set_xy0)
    """numpy.array(dtype=int): The xy-coordinate of the primary corner of the pin."""

    def get_xy1(self):
        """numpy.ndarray(dtype=int): gets the secondary corner of the instance."""
        return self.xy[1]

    def set_xy1(self, value):
        """numpy.ndarray(dtype=int): sets the primary corner of the instance."""
        self.xy[1] = np.asarray(value)
        self.update_pointers()

    xy1 = property(get_xy1, set_xy1)
    """numpy.array(dtype=int): The xy-coordinate of the secondary corner of the rect."""

    @property
    def height(self):
        """int: the height of the rect."""
        return abs(self.xy0[1] - self.xy1[1])

    @property
    def width(self):
        """int: the width of the rect."""
        return abs(self.xy0[0] - self.xy1[0])

    @property
    def size(self):
        """numpy.ndarray(dtype=int): the size of the rect."""
        return np.array([self.width, self.height])

    @property
    def bbox(self):
        """numpy.ndarray(dtype=int): the bounding box of the object. Its default format is
        [[x_ll, y_ll], [x_ur, y_ur]]"""
        return np.sort(np.array([self.xy0, self.xy1]), axis=0)

    def __init__(self, name, xy, layer, netname=None, master=None, params=None, elements=None):
        """
        Constructor.

        Parameters
        ----------
        name : str
            The name of the object.
        xy : numpy.ndarray(dtype=int)
            The xy-coordinates of the object. Its default format is [[x0, y0], [x1, y1]].
        layer : list(str)
            Layer information. Its default format is [layer, purpose].
        netname : str, optional
            The name of net associated with the object. If None, name is used for the net name.
        master : Instance, optional
            Master instance handle.
        params : dict, optional
            Additional parameters of the object.
        """
        self.layer = layer
        if netname is None:
            netname = name
        self.netname = netname
        self.elements = elements
        self.master = master

        LayoutIterableObject.__init__(self, name, xy, params=params, elements=elements)

    def summarize(self):
        """Returns the summary of the object information."""
        return LayoutIterableObject.summarize(self) + ", " + \
                                                      "layer: " + str(self.layer) + ", " + \
                                                      "netname: " + str(self.netname) + ", " + \
                                                      "master: " + str(self.master)


class Text(LayoutObject):
    """
    Text object.

    See Also
    --------
    LayoutObject
    """
    layer = None
    """list(str): the layer information of the text. Its default format is [name, purpose]."""
    text = None
    """str: the text body."""

    def __init__(self, name, xy, layer, text, params=None):
        """
        Constructor.

        Parameters
        ----------
        name : str
            The name of the object.
        xy : numpy.ndarray(dtype=int)
            The xy-coordinates of the object. Its default format is [x, y].
        layer : list(str)
            The layer information of the text. Its default format is [name, purpose].
        text : str
            The text entry.
        """
        self.layer = layer
        self.text = text

        LayoutObject.__init__(self, name, xy, params=params)

    def summarize(self):
        """Returns the summary of the object information."""
        return LayoutObject.summarize(self) + ", " + \
                                              "layer: " + str(self.layer) + ", " + \
                                              "text: " + str(self.text)


class Instance(LayoutIterableObject):
    """
    Instance object, corresponding to a single/mosaic layout instance.

    See Also
    --------
    LayoutIterableObject
    """
    shape = None
    """np.array([int, int]) or None: the shape of the instance mosaic. None if the instance is non-mosaic."""
    _pitch = None
    """np.array([int, int]) or None: the internal variable for pitch."""
    _unit_size = None
    """np.array([int, int]) or None: the internal variable for unit_size."""
    transform = 'R0'
    """str: the transform parameter of the instance."""
    template = None
    """TemplateObject.TemplateObject: the master template of the instance."""
    pins = None
    """dict(): pins of the instance."""
    pointers = dict()
    """dict(): pointers of the instance."""

    # frequently used pointers
    left = None
    right = None
    top = None
    bottom = None
    bottom_left = None
    bottom_right = None
    top_left = None
    top_right = None

    @property
    def xy0(self):
        """numpy.ndarray(dtype=[int, int]): the xy-coordinate of the object."""
        return self.xy

    @property
    def xy1(self):
        """numpy.ndarray(dtype=[int, int]): the secondary xy-coordinate of the object."""
        if self.size is None:
            return self.xy
        else:
            return self.xy + np.dot(self.size, tf.Mt(self.transform).T)

    def get_unit_size(self):
        """numpy.ndarray(dtype=[int, int]): gets the unit size of the object."""
        if self._unit_size is None:
            return self.template.size(params = self.params)
        else:
            return self._unit_size()

    def set_unit_size(self, value):
        """numpy.ndarray(dtype=[int, int]): sets the unit size of the object."""
        self._unit_size = value

    unit_size = property(get_unit_size, set_unit_size)
    """np.array([int, int]) or None: the unit size of the instance. The unit size is the size of the instance with 
    [1, 1] shape."""

    @property
    def size(self):
        """numpy.ndarray(dtype=int): the size of the instance. Its default format is [x_size, y_size]."""
        if self.shape is None:
            return self.unit_size
        else:
            return (self.shape - np.array([1, 1])) * self.pitch + self.unit_size

    def get_pitch(self):
        """numpy.ndarray(dtype=int): gets the pitch of the instance."""
        if self._pitch is None:
            return self.unit_size
        else:
            return self._pitch

    def set_pitch(self, value):
        """numpy.ndarray(dtype=int): sets the pitch of the instance."""
        self._pitch = value

    pitch = property(get_pitch, set_pitch)
    """numpy.ndarray(dtype=int): the pitch of the instance. Its default format is [x_pitch, y_pitch].
    None if template size is used for the instance pitch."""

    def get_spacing(self):
        return self.pitch

    def set_spacing(self, value):
        self.pitch = value

    spacing = property(get_spacing, set_spacing)
    """numpy.ndrarray([int, int]): (deprecated) the pitch of the instance. Previously the pitch was named to spacing,
     to be compatible with GDS-II's notations."""

    @property
    def bbox(self):
        """numpy.ndarray(dtype=int): the bounding box of the instance. Its default format is
        [[x_ll, y_ll], [x_ur, y_ur]]"""
        bbox = np.array([self.xy, self.xy + np.dot(self.size, tf.Mt(self.transform).T)])
        return np.sort(bbox, axis=0)

    def __init__(self, name, xy, template, shape=None, pitch=None, transform='R0', params=None):
        """
        Constructor.

        Parameters
        ----------
        name : str
            The name of the object.
        xy : numpy.ndarray(dtype=int)
            The xy-coordinate of the object. The format is [x0, y0].
        template: Template
            The template object handle.
        shape : numpy.ndarray(dtype=int) or None
            The size of the instance array. The format is [col, row].
        pitch : numpy.ndarray(dtype=int) or None
            The stride of the instance array. Its format is [x_pitch, y_pitch]. If none, the template size is used for
            the array pitch.
        transform : str
            The transform parameter. Possible values are 'R0', 'R90', 'R180', 'R270', 'MX', 'MY'.
        params : dict
            Additional parameters of the object.
        """
        # Preprocess parameters
        if isinstance(xy, Pointer):
            xy = xy.xy  # convert to xy coordinate.
        xy = np.asarray(xy)
        self.template = template
        if shape is not None:
            _shape = np.asarray(shape)
            if _shape.shape != (2, ):
                raise ValueError('Instance shape should be a (2, ) numpy array or None.')
            self.shape = _shape
        if pitch is not None:
            self.pitch = np.asarray(pitch)
        self.transform = transform

        # Construct an array for elements.
        if shape is None:
            elements = self  # np.array([])
        else:
            _shape = tuple(shape)
            elements = np.zeros(_shape, dtype=np.object)
            # elements = LayoutObjectArray(np.zeros(_shape, dtype=np.object))
            _it = np.nditer(elements, flags=['multi_index', 'refs_ok'])
            while not _it.finished:
                _idx = _it.multi_index
                _xy = xy + np.dot(self.pitch * np.array(_idx), tf.Mt(self.transform).T)
                inst = Instance(name=name, xy=_xy, template=self.template, shape=None, pitch=pitch,
                                transform=self.transform, params=params)
                elements[_idx] = inst
                _it.iternext()

        LayoutIterableObject.__init__(self, name, xy, params=params, elements=elements)

        # Create the pin dictionary
        self.pins = dict()
        pins = template.pins(params=self.params)
        if pins is not None:
            if not isinstance(pins, dict):
                raise ValueError("laygo2.Template.pins() should return a dictionary that contains pin information.")
            else:
                for pn, p in pins.items():
                    if shape is not None:
                        elements = []
                        for i in range(shape[0]):
                            elements.append([])
                            for j in range(shape[1]):
                                _xy = p['xy'] + np.dot(self.pitch * np.array([i, j]), tf.Mt(transform).T)
                                pin = Pin(name=pn, xy=_xy, netname=p['netname'], layer=p['layer'], master=self,
                                          elements=None)  # master uses self instead of self.elements[i, j].
                                elements[i].append(pin)
                        elements = np.array(elements)
                    else:
                        elements = None
                    self.pins[pn] = Pin(name=pn, xy=p['xy'], netname=p['netname'], layer=p['layer'], master=self,
                                        elements=elements)

    def summarize(self):
        """Summarizes object information."""
        return LayoutObject.summarize(self) + ", " + \
                                              "template: " + str(self.template) + ", " + \
                                              "size: " + str(self.size) + ", " + \
                                              "shape: " + str(self.shape) + ", " + \
                                              "pitch: " + str(self.pitch) + ", " + \
                                              "transform: " + str(self.transform) + ", " + \
                                              "pins: " + str(self.pins)


# Test
if __name__ == '__main__':
    print("Rect test")
    rect0 = Rect(name='R0', xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'maxI': 0.005})
    print(rect0)
    print("Wire test")
    wire0 = Wire(name='R0', xy=[[0, 0], [0, 100]], width=10, extension=5, layer=['M1', 'drawing'], netname='net0', params={'maxI': 0.005})
    print(wire0)
    print("Pin test")
    pin0 = Pin(name='P0', xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', master=rect0,
               params={'direction':'input'})
    print(pin0)
    print("Text test")
    text0 = Text(name='T0', xy=[0, 0], layer=['text', 'drawing'], text='test', params=None)
    print(text0)
    print("Pointer test")
    pointer0 = Pointer(name='PT0', xy=[0, 0], direction='left', master=rect0)
    print(pointer0)
    print("Template test")
    from templates import Template
    template0_pins = dict()
    template0_pins['in'] = {'xy': [[0, 0], [10, 10]], 'netname': 'in', 'layer': ['M1', 'drawing']}
    template0_pins['out'] = {'xy': [[90, 90], [100, 100]], 'netname': 'out', 'layer': ['M1', 'drawing']}
    template0 = Template(name='my_template0', xy=[[0, 0], [100, 100]], pins=template0_pins)
    print(template0)
    print("Instance test")
    inst0 = Instance(name='I0', xy=[100, 100], template=template0, shape=[3, 2], pitch=[100, 100], transform='R0')
    print(inst0)
    print(inst0.shape)
    for idx, it in inst0.ndenumerate():
        print(idx, it.pins['in'])
    for idx, it in inst0.pins['in'].ndenumerate():
        print(idx, it)
